/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 * 
 * This file is part of de.bsvrz.ars.ars.
 * 
 * de.bsvrz.ars.ars is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.ars.ars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.ars.ars; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

package de.bsvrz.ars.ars.mgmt;

import de.bsvrz.ars.ars.mgmt.datatree.DIdNodeNotFoundException;
import de.bsvrz.ars.ars.mgmt.datatree.DataIdentNode;
import de.bsvrz.ars.ars.mgmt.datatree.DataIdentTree;
import de.bsvrz.ars.ars.mgmt.simulation.SimulationResultData;
import de.bsvrz.ars.ars.mgmt.tasks.*;
import de.bsvrz.ars.ars.mgmt.tasks.tasklistener.RingOfDeathListener;
import de.bsvrz.ars.ars.persistence.CacheManager;
import de.bsvrz.dav.daf.main.ClientReceiverInterface;
import de.bsvrz.dav.daf.main.ClientSenderInterface;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.DataNotSubscribedException;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.SendSubscriptionNotConfirmed;
import de.bsvrz.dav.daf.main.archive.ArchiveData;
import de.bsvrz.dav.daf.main.archive.ArchiveQueryPriority;
import de.bsvrz.dav.daf.main.config.ConfigurationException;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.sys.funclib.dataSerializer.Deserializer;
import de.bsvrz.sys.funclib.dataSerializer.NoSuchVersionException;
import de.bsvrz.sys.funclib.dataSerializer.SerializingFactory;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.losb.datk.MessageType;
import de.bsvrz.sys.funclib.losb.datk.PidQuery;
import de.bsvrz.sys.funclib.losb.ringbuffer.ArchiveRingBuffer;
import de.bsvrz.sys.funclib.losb.ringbuffer.RingBuffer;
import de.bsvrz.sys.funclib.losb.util.JavaSpy;
import de.bsvrz.sys.funclib.losb.util.Util;
import de.bsvrz.sys.funclib.operatingMessage.MessageGrade;
import de.bsvrz.sys.funclib.operatingMessage.MessageSender;
import de.bsvrz.sys.funclib.operatingMessage.MessageState;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.concurrent.Semaphore;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.concurrent.atomic.AtomicLong;

/**
 * Warteschlangen-Manager. Verteilt reinkommende Datentelegramm mit Datensaetzen/Archivanfragen/Archiv-Informationsanfragen/
 * Backup-/Restore-/HeaderRestore-/LZV/Delete-/RequestGap-Auftrgen auf die jeweilige Warteschlange.
 *
 * @author beck et al. projects GmbH
 * @author Alexander Schmidt
 * @version $Revision: 11353 $ / $Date: 2013-06-25 16:12:23 +0200 (Di, 25 Jun 2013) $ / ($Author: jh $)
 */
public class InQueuesMgr implements RingOfDeathListener {

	private static int ID_ARCH_QUEUES_ONLINE = 0, ID_ARCH_QUEUES_REQ = 1, ID_NONARCH_QUEUES = 2;

	// Task Manager
	private static final int TM_QRYA_IDX = 0; /* Archivanfragen Prio A */

	private static final int TM_QRYB_IDX = 1; /* Archivanfragen Prio B */

	private static final int TM_QRYC_IDX = 2; /* Archivanfragen Prio C */

	private static final int TM_ARCINF_IDX = 3; /* Archiv-Informationsanfragen */

	// Anzahl der Query-Tasks pro Taskmanager:

	private static final int TM_QRY_A_NUM_DEFAULT = 3;  // Defaultwert gemaess Datenkatalog

	private static final int TM_QRY_B_NUM_DEFAULT = 6;  // Defaultwert gemaess Datenkatalog

	private static final int TM_QRY_C_NUM_DEFAULT = 9;  // Defaultwert gemaess Datenkatalog

	private static final int TM_ARC_INF_NUM = 4;

	private static final int NUM_OF_MULTITASK_MGR = 4;

	// Single Tasks
	public static final int ST_ARC_SETT_IDX = 0; /* Archiv-Einstellungen */

	public static final int ST_BACKUP_IDX = 1; /* Sicherung */

	public static final int ST_RESTORE_IDX = 2; /* Wiederherstellen */

	public static final int ST_HDR_REST_IDX = 3; /* Verwaltungs-Info-Abgleich */

	public static final int ST_LZV_IDX = 4; /* Lschzeit-nderung */

	public static final int ST_DELETE_IDX = 5; /* Lschen */

	public static final int ST_REQ_GAP_IDX = 6; /* Nachfordern */

	public static final int ST_NUM_QRY_INF_IDX = 7; /* Warteschlangenanfrage */

	public static final int NUM_OF_SINGLETASKS = 8;

	/** Indexe der Eingangswarteschlangen */
	public static final int Q_QRYA_IDX = 0;

	public static final int Q_QRYB_IDX = 1;

	public static final int Q_QRYC_IDX = 2;

	public static final int Q_ARCINF_IDX = 3;

	public static final int Q_BACKUP_IDX = 4;

	public static final int Q_RESTORE_IDX = 5;

	public static final int Q_HDR_REST_IDX = 6;

	public static final int Q_LZV_IDX = 7;

	public static final int Q_DELETE_IDX = 8;

	public static final int Q_REQ_GAP_IDX = 9;

	public static final int Q_ARC_SETT_IDX = 10;

	public static final int Q_SIMPARAM_IDX = 11;

	public static final int Q_NUM_QRY_INF_IDX = 12;

	public static final int NUM_OF_NONARCH_QUEUES = 13;

	private static final int[] CAPA_QUEUES = new int[NUM_OF_NONARCH_QUEUES];



	static {
		CAPA_QUEUES[Q_QRYA_IDX] = 512;
		CAPA_QUEUES[Q_QRYB_IDX] = 512;
		CAPA_QUEUES[Q_QRYC_IDX] = 512;
		CAPA_QUEUES[Q_ARCINF_IDX] = 512;
		CAPA_QUEUES[Q_BACKUP_IDX] = 512;
		CAPA_QUEUES[Q_RESTORE_IDX] = 512;
		CAPA_QUEUES[Q_HDR_REST_IDX] = 512;
		CAPA_QUEUES[Q_LZV_IDX] = 512;
		CAPA_QUEUES[Q_DELETE_IDX] = 512;
		CAPA_QUEUES[Q_REQ_GAP_IDX] = 512;
		CAPA_QUEUES[Q_ARC_SETT_IDX] = 512;
		CAPA_QUEUES[Q_SIMPARAM_IDX] = 999;
		CAPA_QUEUES[Q_NUM_QRY_INF_IDX] = 512;
	}

	/** Zahl der Tasks, die nachgelieferte Datensaetze archivieren */
	public static int NUM_OF_ARCH_QUEUES_REQ = 8;

	// ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	// Parameter fuer die Warteschlangen fuer Online-Daten

	/** Zahl der Tasks, die aktuelle Datensaetze archivieren (sollte Primzahl sein) */
	public static int NUM_OF_ARCH_QUEUES_ONLINE = 61;

	/**
	 * Initiale Maximalgroesse eines Ringpuffers fuer aktuelle Datensaetze. Die Maximalgroesse kann sich auf Kosten anderer Puffer aendern. Mit dynamischen
	 * Puffergroessen kann man Lastspitzen besser abfangen.
	 */
	public static int INITIAL_CAPA_Q_ARCH_TSK_ONLINE = 3000;

	/** Blockgroesse eines Ringpuffers fuer aktuelle Datensaetze. In diesen Bloecken werden die Puffer vergroessert oder verkleinert. */
	private static int CHUNK_Q_ARCH_TSK_ONLINE = 250;

	/** Ab dieser Auslastung werden Non-Write-Tasks angehalten */
	private static double NONWRITE_SUSPEND_LIMIT = 0.8;

	/** Ab dieser Auslastung werden Non-Write-Tasks wieder fortgesetzt (mit Hysterese) */
	private static double NONWRITE_RESUME_LIMIT = 0.7;

	/** Zaehlt die insgesamt in den Warteschlangen enthaltenen Datensaetze */
	private AtomicInteger onlineDataInQueues = new AtomicInteger(0);

	// Ende Parameter
	// ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

	/** Kapazitaet des Ringpuffers fuer nachgelieferte Datensaetze */
	public static int CAPA_Q_ARCH_TSK_REQ = 2000;

	// Alle Ringpuffer:
	private ArchiveRingBuffer[] archiveQueues = new ArchiveRingBuffer[totalArchQueues()];

	private int[] archiveQueuesLeftout = new int[NUM_OF_ARCH_QUEUES_ONLINE];

	@SuppressWarnings({"unchecked"})
	private RingBuffer<ResultData>[] nonArchQueues = new RingBuffer[NUM_OF_NONARCH_QUEUES];

	// Alle direkt referenzierten Tasks (ohne die ueber MultiTaskManager referenzierten Tasks):
	private ArchiveTask[] archiveTasks = new ArchiveTask[totalArchQueues()];

	private SingleTask[] singleTsks = new SingleTask[NUM_OF_SINGLETASKS];

	private MultiTaskManager[] tskMgrs = new MultiTaskManager[NUM_OF_MULTITASK_MGR];

	private FlowControlTask flowCtrl;

	private ArchiveManager archMgr;

	private DataIdentTree didTree;

	private Debug logger = Debug.getLogger();

	private Semaphore terminationSemaphor;

	private boolean terminated = false;

	// hier die Empfaenger-/Senderklassen
	private ArchiveDataReceiver archiveDataReceiver = new ArchiveDataReceiver();

	private QueryReceiver queryReceiver = new QueryReceiver();

	private ArchiveSettingsReceiver arsSettingsReceiver = new ArchiveSettingsReceiver();

	private DataAckSender dataAckSender = new DataAckSender();

	private static AtomicLong TOTAL_DS_RECEIVED = new AtomicLong(0);

	public static final String MSG_PID_QUEUE_FULL = "Warteschlangen-Kapazitaet ueberschritten";

	/** Flag ob eine Ueberlastsituation vorliegt und deswegen alle nicht schreibenden Tasks unterbrochen werden muessen */
	private boolean nonWriteTasksSuspended = false;

	private Object nonWriteTasksSuspendSema = new Object();

	private final static int SHIFT_QUEUES_INTERVAL = 15000;

	private LongTermTaskStatePublisher _longTermTaskStatePublisher;

	private static int totalArchQueues() {
		return NUM_OF_ARCH_QUEUES_ONLINE + NUM_OF_ARCH_QUEUES_REQ;
	}

	/**
	 * Erzeugt den Warteschlangen-Manager.
	 *
	 * @param aMgr  Archiv-Manager
	 * @param dTree DataIdentTree
	 *
	 * @throws ConfigurationException, falls die IDs der Attributgruppe oder des Aspekts der ArchivAnfrageSchnittstelle nicht bestimmt werden koennen.
	 */
	public InQueuesMgr(ArchiveManager aMgr, DataIdentTree dTree) throws ConfigurationException {
		archMgr = aMgr;
		didTree = dTree;
		archMgr.getPersMgr().resetContainerFileMap();

		_longTermTaskStatePublisher = new ArchiveLongTermTaskStatePublisher(archMgr, "Zustandsnderung eines Hintergrundprozesses:", "Zustnde der Hintergrundprozesse:");

		// Queues (mit fester Groesse):
		for(int i = 0; i < NUM_OF_NONARCH_QUEUES; i++) {
			nonArchQueues[i] = new RingBuffer<ResultData>(CAPA_QUEUES[i], CAPA_QUEUES[i]);
		}

		// Queue fuer Simulations-Parametrierung mit ArchivConfig verbinden:
		archMgr.getArchivConfig().setSimConfigQueue(nonArchQueues[Q_SIMPARAM_IDX]);

		// SingleTasks:
		singleTsks[ST_ARC_SETT_IDX] = new ArchiveSettingsTask(archMgr, this, nonArchQueues[Q_ARC_SETT_IDX]);
		singleTsks[ST_BACKUP_IDX] = new BackupTask(archMgr, this, nonArchQueues[Q_BACKUP_IDX], _longTermTaskStatePublisher);
		singleTsks[ST_RESTORE_IDX] = new RestoreTask(archMgr, this, nonArchQueues[Q_RESTORE_IDX], _longTermTaskStatePublisher);
		singleTsks[ST_HDR_REST_IDX] = new HeaderRestoreTask(archMgr, this, nonArchQueues[Q_HDR_REST_IDX]);
		singleTsks[ST_LZV_IDX] = new LZVTask(archMgr, this, nonArchQueues[Q_LZV_IDX], _longTermTaskStatePublisher);
		singleTsks[ST_DELETE_IDX] = new DeleteSimVarTask(archMgr, this, nonArchQueues[Q_DELETE_IDX]);
		singleTsks[ST_REQ_GAP_IDX] = new RequestGapTask(archMgr, this, nonArchQueues[Q_REQ_GAP_IDX]);
		singleTsks[ST_NUM_QRY_INF_IDX] = new QueryInfoTask(archMgr, this, nonArchQueues[Q_NUM_QRY_INF_IDX]);

		// Task-Manager (Multitasks fuer Archivinfo und Queries, parametrierbare Anzahl):
		tskMgrs[TM_QRYA_IDX] = new MultiTaskManager(archMgr, this, nonArchQueues[Q_QRYA_IDX], "QueryPrioA");
		tskMgrs[TM_QRYB_IDX] = new MultiTaskManager(archMgr, this, nonArchQueues[Q_QRYB_IDX], "QueryPrioB");
		tskMgrs[TM_QRYC_IDX] = new MultiTaskManager(archMgr, this, nonArchQueues[Q_QRYC_IDX], "QueryPrioC");
		tskMgrs[TM_ARCINF_IDX] = new MultiTaskManager(archMgr, this, nonArchQueues[Q_ARCINF_IDX], "ArchivInfo");

		flowCtrl = new FlowControlTask(archMgr);

		// parametrierbare Anzahl Query-Tasks starten:
		setQueryTaskNumbers(TM_QRY_A_NUM_DEFAULT, TM_QRY_B_NUM_DEFAULT, TM_QRY_C_NUM_DEFAULT);

		// die Zahl der ArchivInfo-Tasks ist fest:
		for(int i = 0; i < TM_ARC_INF_NUM; i++) {
			tskMgrs[TM_ARCINF_IDX].addTask(new ArchiveInfoQueryTask(aMgr, tskMgrs[TM_ARCINF_IDX]));
		}

		// Queues + Tasks fr Archivierung der aktuellen Datensaetze (fixe, experimentell festgestellte Anzahl):
		for(int i = 0; i < NUM_OF_ARCH_QUEUES_ONLINE; i++) {
			archiveQueues[i] = new ArchiveRingBuffer(CHUNK_Q_ARCH_TSK_ONLINE, INITIAL_CAPA_Q_ARCH_TSK_ONLINE);
			archiveTasks[i] = new ArchiveTask(archMgr, this, i, archiveQueues[i]);
		}
		// Queues + Tasks fr Archivierung der nachgelieferten Datensaetze (fixe Anzahl):
		for(int i = 0; i < NUM_OF_ARCH_QUEUES_REQ; i++) {
			archiveQueues[NUM_OF_ARCH_QUEUES_ONLINE + i] = new ArchiveRingBuffer(CAPA_Q_ARCH_TSK_REQ, CAPA_Q_ARCH_TSK_REQ);
			archiveTasks[NUM_OF_ARCH_QUEUES_ONLINE + i] = new ArchiveRequestedTask(archMgr, this, i, archiveQueues[NUM_OF_ARCH_QUEUES_ONLINE + i]);
		}
	}

	/**
	 * Vergroessert eine Queue (auf Kosten einer anderen, falls die Gesamtgroesse ueberschritten ist).
	 *
	 * @param queue Zu vergroessernde Queue
	 */
	private void enlargeQueue(int queue) {
		int selectedIndex = -1;
		int maxFreeSpace = 0;
		int maxCapacity = 0;
		for(int idx = 0; idx < NUM_OF_ARCH_QUEUES_ONLINE; idx++) {
			if(idx == queue) continue;
			int capacity = archiveQueues[idx].maxSize();
			int freeSpace = capacity - archiveQueues[idx].size();
			if((freeSpace > maxFreeSpace) || ((freeSpace == maxFreeSpace) && (capacity > maxCapacity))) {
				maxCapacity = capacity;
				maxFreeSpace = freeSpace;
				selectedIndex = idx;
			}
		}
		if(selectedIndex >= 0 && selectedIndex != queue) {
			if(archiveQueues[selectedIndex].maxSize() - archiveQueues[selectedIndex].size() >= 2 * CHUNK_Q_ARCH_TSK_ONLINE) {
				archiveQueues[selectedIndex].changeMaxSize(-CHUNK_Q_ARCH_TSK_ONLINE);
				archiveQueues[queue].changeMaxSize(CHUNK_Q_ARCH_TSK_ONLINE);
			}
			else {
				logger.warning("Zum Verkleinern ausgewhlte ArchivierungsQueue (index: " + selectedIndex + ") konnte nicht verkleinert werden werden.");
			}
		}
		else {
			logger.warning("Es konnte keine ArchivierungsQueue zum Verkleinern ausgewhlt werden.");
		}
	}

	/**
	 * Setzt die Anzahl der Threads fuer Archivanfragen auf die angegebenen Werte. Falls Threads entfernt werden, werden noch behandelte Anfragen zuende
	 * bearbeitet.
	 *
	 * @param numHi  Anzahl Tasks der Prioritaet "hoch"
	 * @param numMid Anzahl Tasks der Prioritaet "mittel"
	 * @param numLo  Anzahl Tasks der Prioritaet "niedrig"
	 */
	public void setQueryTaskNumbers(int numHi, int numMid, int numLo) {
		installQueryTask(tskMgrs[TM_QRYA_IDX], numHi);
		installQueryTask(tskMgrs[TM_QRYB_IDX], numMid);
		installQueryTask(tskMgrs[TM_QRYC_IDX], numLo);
	}

	public int getHiQueryTaskNum() {
		return tskMgrs[TM_QRYA_IDX].getTaskNum();
	}

	public int getMidQueryTaskNum() {
		return tskMgrs[TM_QRYB_IDX].getTaskNum();
	}

	public int getLoQueryTaskNum() {
		return tskMgrs[TM_QRYC_IDX].getTaskNum();
	}

	private void installQueryTask(MultiTaskManager mtm, int anz) {
		if(mtm.isTaskTerminated()) return;

		int anzQTasks = mtm.getTaskNum();
		if(anz > anzQTasks) {
			for(int i = 0; i < (anz - anzQTasks); i++) {
				ArchiveQueryTask newTask = new ArchiveQueryTask(archMgr, mtm, flowCtrl);
				newTask.setMaximumRequestsPerApplication(archMgr.getMaximumQueriesPerApplication());
				mtm.addTask(newTask);
			}
		}
		else {
			for(int i = 0; i < (anzQTasks - anz); i++) {
				mtm.removeTask();
			}
		}
	}

	/**
	 * Prueft ob Tasks temporaer blockiert werden sollen und blockiert ggf. den aufrufenden Task.
	 * <p/>
	 * Diese Methode und {@link #suspendNonWriteTasks(boolean)} verwenden eine reduzierte Synchronisierung um die Zahl der durchlaufenen synchronized-Bloecke
	 * gering zu halten.
	 *
	 * @throws InterruptedException
	 */
	public void checkForSuspendNonWriteTask() throws InterruptedException {
		if(nonWriteTasksSuspended)	// synchronized nur ausfuehren falls notwendig
		{
			synchronized(nonWriteTasksSuspendSema) {
				if(nonWriteTasksSuspended)	// hier ist der echte Test
				{
					nonWriteTasksSuspendSema.wait();
				}
			}
		}
	}

	/**
	 * Setzt das Suspend-Flag oder hebt es auf abhaengig von den beiden Grenzen.
	 * <p/>
	 * Diese Methode und {@link #checkForSuspendNonWriteTask()} verwenden eine reduzierte Synchronisierung um die Zahl der durchlaufenen synchronized-Bloecke
	 * gering zu halten. Voraussetzung: diese Methode wird nur von einem Thread aufgerufen.
	 */
	private void suspendNonWriteTasks() {
		// Zugriff auf nonWriteTasksSuspended findet nur in dieser Methode statt
		if(!nonWriteTasksSuspended && getQueueLoad() > NONWRITE_SUSPEND_LIMIT) {
			nonWriteTasksSuspended = true; // zu beliebigen Zeiten moeglich
			logger.fine("Bearbeitung von Archivanfragen wegen Ueberlast ausgesetzt");
			//Absetzen der Betriebsmeldung
			MessageSender.getInstance().sendMessage(
					de.bsvrz.sys.funclib.operatingMessage.MessageType.APPLICATION_DOMAIN,
					MessageGrade.INFORMATION,
					"Die Warteschlangen sind zu " + (NONWRITE_SUSPEND_LIMIT * 100) + "% gefllt - das Schreiben von Daten erhlt eine hhere Prioritt."
			);
		}
		else if(nonWriteTasksSuspended && getQueueLoad() < NONWRITE_RESUME_LIMIT) {
			synchronized(nonWriteTasksSuspendSema) {
				nonWriteTasksSuspended = false;	// nur synchronisiert moeglich
				nonWriteTasksSuspendSema.notifyAll();
				logger.fine("Bearbeitung von Archivanfragen wieder aufgenommen");
				//Absetzen der Betriebsmeldung
				MessageSender.getInstance().sendMessage(
						de.bsvrz.sys.funclib.operatingMessage.MessageType.APPLICATION_DOMAIN,
						MessageGrade.INFORMATION,
						"Die Warteschlangen sind nur noch zu " + (NONWRITE_RESUME_LIMIT * 100) + "% gefllt. Alle Tasks arbeiten wieder mit gleicher Prioritt."
				);
			}
		}
	}

	/**
	 * In dieser Methode wird allen Tasks (falls notwendig) die Moeglichkeit gegeben, benoetigte Objekte vom DAV zu laden.
	 *
	 * @throws ConfigurationException
	 */
	public void getObjectsFromDav() throws ConfigurationException {
		RestoreTask.getObjectsFromDAV(archMgr.getDavCon());
		HeaderRestoreTask.getObjectsFromDAV(archMgr.getDavCon());
		LZVTask.getObjectsFromDAV(archMgr.getDavCon());
		DeleteSimVarTask.getObjectsFromDAV(archMgr.getDavCon());
		ArchiveQueryTask.getObjectsFromDAV(archMgr.getDavCon());
		ArchiveInfoQueryTask.getObjectsFromDAV(archMgr.getDavCon());
	}

	/**
	 * In dieser Methode wird allen Tasks (falls notwendig) die Moeglichkeit gegeben, sich auf Objekte anzumelden. Abmelden geschieht zentral ueber den {@link
	 * de.bsvrz.sys.funclib.losb.kernsoftware.ConnectionManager}.
	 */
	public void subscribeObjects() {
		ArchiveSettingsTask.subscribeObjects(archMgr, arsSettingsReceiver);
		ArchiveQueryTask.subscribeObjects(archMgr, queryReceiver);
	}

	/** Startet alle Tasks (Multi- und Single-Tasks) sowie die Archiv-Tasks. */
	public void startAllTasks() {
		_longTermTaskStatePublisher.setDavConnection(archMgr.getDavCon());
		resetDSCounter();
		ArchiveTask.resetDSCounter();
		ArchiveRequestedTask.resetDSCounter();
		for(int i = 0; i < NUM_OF_MULTITASK_MGR; i++) {
			tskMgrs[i].start();
		}
		for(int i = 0; i < NUM_OF_SINGLETASKS; i++) {
			singleTsks[i].start();
		}
		for(int i = 0; i < totalArchQueues(); i++) {
			archiveTasks[i].start();
		}
	}

	/**
	 * Beendet alle Tasks (Multi- und Single-Tasks) sowie die Archiv-Tasks.
	 *
	 * @throws InterruptedException
	 */
	public void stopAllTasks() throws InterruptedException {
		terminated = true;
		int totalTasks = totalArchQueues() + NUM_OF_MULTITASK_MGR + NUM_OF_SINGLETASKS;
		terminationSemaphor = new Semaphore(totalTasks);
		terminationSemaphor.acquire(totalTasks);

		for(int i = 0; i < NUM_OF_SINGLETASKS; i++) {
			if(singleTsks[i].getState() == Thread.State.NEW) {
				ringOfDeath(singleTsks[i]);		// wenn nicht gestartet, Totenglckchen direkt aufrufen
			}
			else {
				singleTsks[i].terminateTask();
			}
		}
		for(int i = 0; i < NUM_OF_MULTITASK_MGR; i++) {
			if(tskMgrs[i].getState() == Thread.State.NEW) {
				ringOfDeath(tskMgrs[i]);		// wenn nicht gestartet, Totenglckchen direkt aufrufen
			}
			else {
				tskMgrs[i].terminateTask();
			}
		}
		waitForEmptyArchiveQueues();
		for(int i = 0; i < totalArchQueues(); i++) {
			if(archiveTasks[i].getState() == Thread.State.NEW) {
				ringOfDeath(archiveTasks[i]);		// wenn nicht gestartet, Totenglckchen direkt aufrufen
			}
			else {
				archiveTasks[i].terminateTask();
			}
		}

		Thread.sleep(10000);
		while(terminationSemaphor.availablePermits() < totalTasks)
		{
			final StringBuilder builder = new StringBuilder();
			final StringBuilder taskBuilder = new StringBuilder();
			int numTasksAlive = 0;

			for(int i = 0; i < NUM_OF_SINGLETASKS; i++) {
				if(singleTsks[i].isAlive()) {
					taskBuilder.append(singleTsks[i].getName());
					taskBuilder.append(",\n");
					numTasksAlive++;
				}
			}
			for(int i = 0; i < NUM_OF_MULTITASK_MGR; i++) {
				if(tskMgrs[i].isAlive()) {
					taskBuilder.append(tskMgrs[i].getName());
					taskBuilder.append(",\n");
					numTasksAlive++;
				}
			}
			for(int i = 0; i < totalArchQueues(); i++) {
				if(archiveTasks[i].isAlive()) {
					taskBuilder.append(archiveTasks[i].getName());
					taskBuilder.append(",\n");
					numTasksAlive++;
				}
			}

			if(numTasksAlive == 0) break;
			builder.append("Tasks werden beendet...\n");
			builder.append("Es mssen noch ");
			builder.append(numTasksAlive);
			builder.append(" von ");
			builder.append(totalTasks);
			builder.append(" Tasks beendet werden.\n");
			builder.append("Verbleibende Tasks:\n");
			builder.append(taskBuilder.substring(0, taskBuilder.length() - 2));
			logger.info(builder.toString());
			Thread.sleep(10000);
		}

		terminationSemaphor.acquire(totalTasks);

		for(int i = 0; i < NUM_OF_SINGLETASKS; i++) {
			singleTsks[i] = null;
		}
		for(int i = 0; i < NUM_OF_MULTITASK_MGR; i++) {
			tskMgrs[i] = null;
		}
		for(int i = 0; i < totalArchQueues(); i++) {
			archiveTasks[i] = null;
		}

		archiveQueues = null;
		nonArchQueues = null;
		logger.info("Alle Tasks korrekt beendet");
	}

	private void waitForEmptyArchiveQueues() {
		int numOfWaitingData = 0;
		for(int i = 0; i < totalArchQueues(); i++) {
			numOfWaitingData += archiveQueues[i].size();
		}

		if(numOfWaitingData > 0) {
			logger.info(numOfWaitingData + " Datensaetze in der Warteschlange noch zu archivieren...");
			try {
				while(true) {
					numOfWaitingData = 0;
					for(int i = 0; i < totalArchQueues(); i++) {
						numOfWaitingData += archiveQueues[i].size();
					}
					if(numOfWaitingData == 0) break;
					Thread.sleep(50);
				}
			}
			catch(InterruptedException e) {
				logger.error("Warten auf leere Archiv-Warteschlange unterbrochen.");
			}
			logger.info("Alle Datensaetze archiviert.");
		}
	}

	/* (non-Javadoc)
		 * @see de.bsvrz.ars.ars.mgmt.tasks.RingOfDeathListener#ringOfDeath(de.bsvrz.ars.ars.mgmt.tasks.Task)
		 */
	public synchronized void ringOfDeath(Task doomedTask) {
		logger.finer("Task beendet: " + doomedTask.getName());
		terminationSemaphor.release();
	}

	private static int calcArchivTaskIndex(long objID, long atgID, long aspID, int simVar, int numOfQueues) {
		return ((int)(objID + (atgID * 7) + (aspID * 17) + (simVar * 5)) & 0x7fffffff) % numOfQueues;
	}

	/**
	 * Verteilung der Datenidentifikationen auf die Online-Archiv-Tasks erfolgt anhand der IDs. Jede Datenidentifikation wird somit immer in dieselbe ArchiveQueue
	 * eingefuegt. Auf diese Weise kann die Reihenfolge des Eintreffens wesentlich einfacher erhalten werden.
	 *
	 * @param objID  Objekt-ID
	 * @param atgID  Attributgruppen-ID
	 * @param aspID  Aspekt-ID
	 * @param simVar Simulationsvariante
	 *
	 * @return Summe der IDs und simVar modulo Anzahl der ArchivTasks
	 */
	public static int calcOnlineArchivTaskIndex(long objID, long atgID, long aspID, int simVar) {
		return calcArchivTaskIndex(objID, atgID, aspID, simVar, NUM_OF_ARCH_QUEUES_ONLINE);
	}

	/**
	 * Verteilung der Datenidentifikationen auf die Nachgefordert-Archiv-Tasks erfolgt anhand der IDs. Jede Datenidentifikation wird somit immer in dieselbe
	 * ArchiveQueue eingefuegt. Auf diese Weise kann die Reihenfolge des Eintreffens wesentlich einfacher erhalten werden.
	 *
	 * @param objID  Objekt-ID
	 * @param atgID  Attributgruppen-ID
	 * @param aspID  Aspekt-ID
	 * @param simVar Simulationsvariante
	 *
	 * @return Summe der IDs und simVar modulo Anzahl der ArchivTasks
	 */
	public static int calcReqArchivTaskIndex(long objID, long atgID, long aspID, int simVar) {
		return calcArchivTaskIndex(objID, atgID, aspID, simVar, NUM_OF_ARCH_QUEUES_REQ);
	}

	/**
	 * Verteilung der Datenidentifikationen auf die Archiv-Tasks erfolgt anhand der IDs. Jede Datenidentifikation wird somit immer in dieselbe ArchiveQueue
	 * eingefuegt. Auf diese Weise kann die Reihenfolge des Eintreffens wesentlich einfacher erhalten werden.
	 *
	 * @param rd Ergebnisdatensatz
	 *
	 * @return Summe der IDs modulo Anzahl der ArchivTasks
	 */
	private int calcOnlineArchivTaskIndex(ResultData rd) {
		DataDescription dd = rd.getDataDescription();
		return calcOnlineArchivTaskIndex(
				rd.getObject().getId(), dd.getAttributeGroup().getId(), dd.getAspect().getId(), dd.getSimulationVariant()
		);
	}

	/**
	 * Fuegt einen nachgeforderten Datensatz in die entsprechende Queue ein.
	 *
	 * @param aTime Archivzeit
	 * @param rd    Datensatz
	 */
	public void insertInArchiveQueueReq(long archiveTime, ArchiveData ad) {
		DataDescription dd = ad.getDataDescription();
		int qIdx = calcReqArchivTaskIndex(
				ad.getObject().getId(), dd.getAttributeGroup().getId(), dd.getAspect().getId(), dd.getSimulationVariant()
		);
		boolean pushSuccessful = false;
		long nextTimeoutTime = archiveTime + SHIFT_QUEUES_INTERVAL;
		try {
			while(!(pushSuccessful = archiveQueues[NUM_OF_ARCH_QUEUES_ONLINE + qIdx].push(ad, archiveTime, 5000))) {
				nextTimeoutTime = managePushTimeout(nextTimeoutTime, "nachgeforderter Datensatz", -1);
			}
		}
		catch(InterruptedException e) {
			// ignored
		}

		if(pushSuccessful) {
			ArchiveRequestedTask.incrementTotalDS_Queued();
		}
		else {
			logCapaExceeded(archiveQueues[NUM_OF_ARCH_QUEUES_ONLINE + qIdx].maxSize(), "Nachforderung", null);
		}
	}

	/**
	 * Fgt dem Simulationsvarianten-Lsch-Task einen Auftrag hinzu.
	 *
	 * @param resultData Auftrag.
	 *
	 * @return <code>true</code> falls der Auftrag eingefgt werden konnte. <code>false</code> sonst.
	 */
	public boolean insertInSimVarDeleteQueue(ResultData resultData) {
		boolean result = nonArchQueues[Q_DELETE_IDX].push(resultData);
		if(!result) logCapaExceeded(nonArchQueues[Q_DELETE_IDX].maxSize(), "Loeschen SimVar", resultData);
		return result;
	}

	/**
	 * Fgt dem Parametrierungs-Task einen weiteren Auftrag hinzu.
	 *
	 * @param resultData Auftrag. Enthlt Informationen, auf welche Datenidentifikationen sich das Archivsystem anmelden soll.
	 *
	 * @return	<code>true</code> falls der Auftrag eingefgt werden konnte. <code>false</code> sonst.
	 */
	public boolean insertSimVarParam(SimulationResultData resultData) {
		boolean result = nonArchQueues[Q_SIMPARAM_IDX].push(resultData);
		if(!result) {
			logCapaExceeded(nonArchQueues[Q_SIMPARAM_IDX].maxSize(), "SimVar Parametrierung ndern", resultData);
		}
		else {
			synchronized(archMgr.getArchivConfig())		//Archivconfig aufwecken
			{
				archMgr.getArchivConfig().notifyAll();
			}
		}
		return result;
	}

	/**
	 * Fgt dem Lsch-Task (automatisches Lschen) einen Auftrag hinzu.
	 *
	 * @param resultData Auftrag.
	 *
	 * @return <code>true</code> falls der Auftrag eingefgt werden konnte. <code>false</code> sonst.
	 */
	public boolean insertInDeleteRegularQueue(ResultData resultData) {
		boolean result = nonArchQueues[Q_LZV_IDX].push(resultData);
		if(!result) logCapaExceeded(nonArchQueues[Q_LZV_IDX].maxSize(), "Loeschen", resultData);
		return result;
	}

	/**
	 * Fgt dem Nachfordern-Task einen Auftrag hinzu.
	 *
	 * @param resultData Auftrag.
	 *
	 * @return <code>true</code> falls der Auftrag eingefgt werden konnte. <code>false</code> sonst.
	 */
	public boolean insertInRequestQueue(ResultData resultData) {
		boolean result = nonArchQueues[Q_REQ_GAP_IDX].push(resultData);
		if(!result) logCapaExceeded(nonArchQueues[Q_REQ_GAP_IDX].maxSize(), "Nachfordern", resultData);
		return result;
	}

	/**
	 * Fgt dem LZV-Task einen Auftrag hinzu.
	 *
	 * @param resultData Auftrag.
	 *
	 * @return <code>true</code> falls der Auftrag eingefgt werden konnte. <code>false</code> sonst.
	 */
	public boolean insertInLZVQueue(ResultData resultData) {
		boolean result = nonArchQueues[Q_LZV_IDX].push(resultData);
		if(!result) logCapaExceeded(nonArchQueues[Q_LZV_IDX].maxSize(), "LZV", resultData);
		return result;
	}


	/**
	 * Fgt dem Sicherungs-Task einen Auftrag hinzu.
	 *
	 * @param resultData Auftrag.
	 *
	 * @return <code>true</code> falls der Auftrag eingefgt werden konnte. <code>false</code> sonst.
	 */
	public boolean insertInBackupQueue(ResultData resultData) {
		boolean result = nonArchQueues[Q_BACKUP_IDX].push(resultData);
		if(!result) logCapaExceeded(nonArchQueues[Q_BACKUP_IDX].maxSize(), "Sicherung", resultData);
		return result;
	}

	/**
	 * Fuegt einen nachgelieferten Datensatz in die entsprechende Queue ein.
	 *
	 * @param aTime Archivzeit
	 * @param rd    Datensatz
	 */
	public void insertCloseContainer(long objID, ArchiveTask.CloseContainerObject cco) {
		int tskIdx = calcOnlineArchivTaskIndex(objID, cco.atgID, cco.aspID, 0);
		boolean pushSuccessful = false;
		long nextTimeoutTime = System.currentTimeMillis() + SHIFT_QUEUES_INTERVAL;
		try {
			while(!(pushSuccessful = archiveQueues[tskIdx].push(cco, objID, 5000))) {
				nextTimeoutTime = managePushTimeout(nextTimeoutTime, "CloseContainer-Datensatz", tskIdx);
			}
		}
		catch(InterruptedException e) {
			// ignored
		}
		accountPushResult(pushSuccessful, tskIdx, "Close-Container-Objekt", null);
	}

	private long managePushTimeout(long nextTimeoutTime, final String usageText, final int tskIdx) {
		final long now = System.currentTimeMillis();
		// Nur bis zu einer Gesamtauslastung von 83% sind bei 3000 Datenstzen pro Queue durchschnittlich noch 2 Chunks mit je 250 Datenstzen frei
		if(now >= nextTimeoutTime) {
			nextTimeoutTime = now + SHIFT_QUEUES_INTERVAL;
			final double load = getQueueLoad();
			if(tskIdx >= 0) {
				if(load <= 0.83) {
					logger.warning(
							"Ein " + usageText + " kann wegen einer vollen Archiv-Queue nicht weitergegeben werden. Die Auslastung aller "
							+ "Queues zusammen betrgt " + load * 100 + "%. Es wird versucht die Kapazitt der betroffenen Queue auf Kosten "
							+ "einer anderen Queue zu vergrern. Anschlieend wird erneut versucht den betroffenen Datensatz weiterzuleiten."
					);
					enlargeQueue(tskIdx);
				}
				else {
					logger.warning(
							"Archivsystem blockiert, weil ein " + usageText + " wegen einer vollen Archiv-Queue nicht weitergegeben werden "
							+ "kann. Die Auslastung aller Queues zusammen betrgt " + load * 100 + "%."
					);
				}
			}
			else {
				logger.warning(
						"Archivsystem blockiert, weil ein " + usageText + " wegen einer vollen Archiv-Queue nicht weitergegeben werden kann."
				);
			}
		}
		return nextTimeoutTime;
	}

	private void accountPushResult(final boolean pushSuccessful, final int tskIdx, final String logTitle, final ResultData resultData) {
		if(pushSuccessful) {
			ArchiveTask.incrementTotalDS_Queued();
			onlineDataInQueues.incrementAndGet();
			if(archiveQueuesLeftout[tskIdx] > 0) {
				logCapaNoLongerExceeded(archiveQueues[tskIdx].maxSize(), logTitle, archiveQueuesLeftout[tskIdx]);
				archiveQueuesLeftout[tskIdx] = 0;
			}
		}
		else {
			// Queues voll, Datensatz verworfen, Logging starten
			if(archiveQueuesLeftout[tskIdx] == 0) {
				logCapaExceeded(archiveQueues[tskIdx].maxSize(), logTitle, resultData);
			}
			archiveQueuesLeftout[tskIdx]++;
		}
	}


	private void logCapaExceeded(int maxsize, String qname, ResultData rd) {
		logger.warning(
				"Kapazitaet (" + maxsize + ") der Warteschlange '" + qname + "' ueberschritten. Datensatz wurde verworfen:" + Debug.NEWLINE + (rd != null
				                                                                                                                               ? Util.rd2Str(rd)
				                                                                                                                               : "")
		);
	}

	/**
	 * Betriebsmeldung absetzen: Warteschlange voll.
	 *
	 * @param maxsize Kapazitt der Warteschlange
	 * @param qname   Name der Warteschlange
	 * @param rd      Datensatz.
	 */
	protected void logSendMsgCapaExceeded(int maxsize, String qname, ResultData rd) {
		String msg =
				"Kapazitaet (" + maxsize + ") der Warteschlange '" + qname + "' ueberschritten. Datensatz wurde verworfen:" + Debug.NEWLINE + Util.rd2Str(rd);
		logger.error(msg);
		MessageSender.getInstance().sendMessage(
				MSG_PID_QUEUE_FULL, de.bsvrz.sys.funclib.operatingMessage.MessageType.APPLICATION_DOMAIN, "", MessageGrade.FATAL, MessageState.MESSAGE, msg
		);
	}

	protected void logCapaNoLongerExceeded(int maxsize, String qname, int leftouts) {
		logger.info("Kapazitaet (" + maxsize + ") der Warteschlange '" + qname + "' wieder ausreichend. " + leftouts + " Datensaetze wurde verworfen.");
	}

	/** @return Status der Warteschlangen fuer aktuelle Datensaetze */
	public float[][] getArchiveQueuesOnlineStatus() {
		return getQueueStatus(ID_ARCH_QUEUES_ONLINE);
	}

	/** @return Status der Warteschlangen fuer nachgelieferte Datensaetze */
	public float[][] getArchiveQueuesRequestedStatus() {
		return getQueueStatus(ID_ARCH_QUEUES_REQ);
	}

	/** @return Status der Warteschlangen fuer eingehende, nicht zu archivierende Objekte */
	public float[][] getNonArchiveQueuesStatus() {
		return getQueueStatus(ID_NONARCH_QUEUES);
	}

	@SuppressWarnings({"unchecked"})
	private float[][] getQueueStatus(int whichQueues) {
		float[][] erg;
		if(whichQueues == ID_ARCH_QUEUES_ONLINE) {
			erg = new float[NUM_OF_ARCH_QUEUES_ONLINE][2];
		}
		else if(whichQueues == ID_ARCH_QUEUES_REQ) {
			erg = new float[NUM_OF_ARCH_QUEUES_REQ][2];
		}
		else if(whichQueues == ID_NONARCH_QUEUES) {
			erg = new float[NUM_OF_NONARCH_QUEUES][2];
		}
		else {
			return new float[][]{};
		}

		RingBuffer queue = null;
		for(int i = 0; i < erg.length; i++) {
			if(whichQueues == ID_ARCH_QUEUES_ONLINE) {
				queue = archiveQueues[i];
			}
			else if(whichQueues == ID_ARCH_QUEUES_REQ) {
				queue = archiveQueues[NUM_OF_ARCH_QUEUES_ONLINE + i];
			}
			else if(whichQueues == ID_NONARCH_QUEUES) queue = nonArchQueues[i];
			int size = queue.size();
			erg[i][0] = size;
			erg[i][1] = (float)size / queue.maxSize();
		}
		return erg;
	}

	/**
	 * @return Zahl aller seit dem Start oder letzten {@link #resetDSCounter()} von {@link #getArchiveDataReceiver()} empfangenen Datensaetze, egal ob diese
	 *         archiviert werden oder nicht. Kann zur Test-Synchronisation verwendet werden.
	 */
	public static long getTotalDSReceived() {
		return TOTAL_DS_RECEIVED.longValue();
	}

	/** @return Gesamtkapazitaet aller Ringpuffers fuer aktuelle Datensaetze */
	public static int getTotalOnlineQueuesCapa() {
		return NUM_OF_ARCH_QUEUES_ONLINE * INITIAL_CAPA_Q_ARCH_TSK_ONLINE;
	}

	/** @return Auslastungsgrad der Warteschlangen. Anhand dieses Wertes wird ermittelt, ob die Bearbeitung nicht schreibender Tasks ausgesetzt wird. */
	public double getQueueLoad() {
		return (double)countOnlineDataInQueues() / (double)getTotalOnlineQueuesCapa();
	}

	public int countOnlineDataInQueues() {
		return onlineDataInQueues.intValue();
	}

	/** Setzt alle Zaehler auf 0. */
	private static void resetDSCounter() {
		TOTAL_DS_RECEIVED.set(0);
	}

	public void decrOnlineDataInQueues() {
		onlineDataInQueues.decrementAndGet();
	}


	/**
	 * Nur fuer Testzwecke: Liefert den gewuenschten ArchiveTask
	 *
	 * @param objId  Objekt-ID
	 * @param atgId  Attributgruppen-ID
	 * @param aspId  Aspekt-ID
	 * @param simVar Simulationsvariante
	 *
	 * @return ArchiveTask
	 *
	 * @see #calcArchivTaskIndex(long,long,long, int, int)
	 */
	public ArchiveTask getOnlineArchiveTask(long objId, long atgId, long aspId, int simVar) {
		return archiveTasks[calcArchivTaskIndex(objId, atgId, aspId, simVar, NUM_OF_ARCH_QUEUES_ONLINE)];
	}

	/**
	 * Liefert den Task, der fuer die Bearbeitung der Archiveinstellungen verantwortlich ist. Dies ist notwendig, damit der {@link ArchiveManager} beim Start einen
	 * {@link de.bsvrz.ars.ars.mgmt.tasks.ArchiveSettingsTask.ArSSettingListener} einhaengen und darauf warten kann, dass die Bearbeitung der Archiveinstellungen
	 * abgeschlossen ist. Das Archivsystem wartet beim Start auf die Archiv-Einstellungen, da es nicht sinnvoll ist, die ersten Sekunden mit Default-Werten
	 * loszulaufen.
	 *
	 * @return Den Task, der fuer die Bearbeitung der Archiveinstellungen verantwortlich ist.
	 */
	public ArchiveSettingsTask getArchiveSettingsTask() {
		return (ArchiveSettingsTask)getSingleTask(InQueuesMgr.ST_ARC_SETT_IDX);
	}


	/**
	 * Liefert den gewuenschten SingleTask (um etwa Listener einzuhaengen)
	 *
	 * @param taskIndex Index der internen SingleTask-Liste
	 *
	 * @return SingleTask
	 */
	public SingleTask getSingleTask(int taskIndex) {
		return singleTsks[taskIndex];
	}

	/**
	 * Die Klasse {@link ArchivConfig} meldet die zu archivierenden Daten mit dem Objekt als Empfaenger an, das von dieser Methode geliefert wird.
	 *
	 * @return Empfaengerobjekt fuer zu archivierende Daten.
	 */
	public ArchiveDataReceiver getArchiveDataReceiver() {
		return archiveDataReceiver;
	}

	/**
	 * Die Klasse {@link ArchivConfig} meldet die Quittungen fuer die zu archivierenden Daten mit dem Objekt als Sender an, das von dieser Methode geliefert wird.
	 *
	 * @return Senderobjekt fuer die Quittungen der zu archivierenden Daten.
	 */
	public DataAckSender getDataAckSender() {
		return dataAckSender;
	}

	/** Empfaengerklasse fuer Archiveinstellungen */
	public class ArchiveSettingsReceiver extends DataReceiver {

		public void processData(ResultData rd) {
			logger.finer("Archiveinstellung erhalten: ", rd);
			if(rd.hasData()) {
				if(!nonArchQueues[Q_ARC_SETT_IDX].push(rd)) logSendMsgCapaExceeded(nonArchQueues[Q_ARC_SETT_IDX].maxSize(), "Archiveinstellungen", rd);
			}
		}
	}

	/** Empfaengerklasse fuer Archivanfragen */
	public class QueryReceiver extends DataReceiver {

		private Deserializer deserializer = null;

		public void processData(ResultData rd) {
			if(rd.hasData()) {
				int msgType = rd.getData().getUnscaledValue(PidQuery.ATT_MESSAGE_TYP_NAME).intValue();
				switch(msgType) {
					case MessageType.QUERY:  // Anfrage
						ArchiveQueryPriority prio;
						try {
							prio = getQueryPrio(rd);
						}
						catch(Exception e) {
							logger.warning("Die Prioritaet einer Archivanfrage konnte nicht ermittelt werden: " + rd, e.getMessage());
							prio = ArchiveQueryPriority.MEDIUM;
						}
						int queueIdx = ArchiveQueryPriority.HIGH == prio ? Q_QRYA_IDX : (ArchiveQueryPriority.MEDIUM == prio ? Q_QRYB_IDX : Q_QRYC_IDX);
						if(!nonArchQueues[queueIdx].push(rd)) logSendMsgCapaExceeded(nonArchQueues[queueIdx].maxSize(), "Archivanfrage " + prio, rd);
						break;
					case MessageType.QUERY_INFO:		  // Informationsanfrage:
						if(!nonArchQueues[Q_ARCINF_IDX].push(rd)) logCapaExceeded(nonArchQueues[Q_ARCINF_IDX].maxSize(), "Archiv-Informationsanfrage ", rd);
						break;
					case MessageType.BACKUP:			  // Sicherungsauftrag:
						if(!nonArchQueues[Q_BACKUP_IDX].push(rd)) logCapaExceeded(nonArchQueues[Q_BACKUP_IDX].maxSize(), "Sicherung", rd);
						break;
					case MessageType.DELETE:			  // Loeschauftrag fuer eine Simulationsvariante:
						if(!nonArchQueues[Q_DELETE_IDX].push(rd)) logCapaExceeded(nonArchQueues[Q_DELETE_IDX].maxSize(), "Loeschen SimVar", rd);
						break;
					case MessageType.DELETION_TIME:	 // Loeschzeitpunkt-Aenderung:
						if(!nonArchQueues[Q_LZV_IDX].push(rd)) logCapaExceeded(nonArchQueues[Q_LZV_IDX].maxSize(), "Loeschzeitpunkt", rd);
						break;
					case MessageType.RESTORE:			// Wiederherstellung:
						if(!nonArchQueues[Q_RESTORE_IDX].push(rd)) logCapaExceeded(nonArchQueues[Q_RESTORE_IDX].maxSize(), "Wiederherstellung", rd);
						break;
					case MessageType.HEADER_RESTORE:	// Verwaltungsinformationsabgleich:
						if(!nonArchQueues[Q_HDR_REST_IDX].push(rd)) logCapaExceeded(nonArchQueues[Q_HDR_REST_IDX].maxSize(), "Verwaltungsinfoabgleich", rd);
						break;
					case MessageType.STREAM_CONTROL:	// Flusskontrolle:
						flowCtrl.processFlowCtrl(rd);
						break;
					case MessageType.REQUEST_DID:
					case MessageType.REQUEST_AUTOM:
						if(!nonArchQueues[Q_REQ_GAP_IDX].push(rd)) logCapaExceeded(nonArchQueues[Q_REQ_GAP_IDX].maxSize(), "Nachforderungsauftrag", rd);
						break;
					case MessageType.NUM_QUERIES_INFO:
						if(!nonArchQueues[Q_NUM_QRY_INF_IDX].push(rd)) logCapaExceeded(nonArchQueues[Q_NUM_QRY_INF_IDX].maxSize(), "Anfrage Warteschlangenkapazitt", rd);
						break;

					default:
						logger.warning("Nachrichtentyp erhalten, der nicht bearbeitet werden kann: " + msgType + " aus: " + Util.rd2Str(rd));
				}
			}
		}

		private ArchiveQueryPriority getQueryPrio(ResultData rd) throws IOException, NoSuchVersionException {
			byte[] data = rd.getData().getUnscaledArray(PidQuery.ATT_DATA_NAME).getByteArray();
			InputStream stream = new ByteArrayInputStream(data, 4, data.length - 4);
			int deserVersion = Util.getSerVersion(data);
			if(deserializer == null || deserializer.getVersion() != deserVersion) {
				deserializer = SerializingFactory.createDeserializer(deserVersion, stream);
			}
			else {
				deserializer.setInputStream(stream);
			}
			return ArchiveQueryPriority.getInstance(deserializer.readInt()); // Prioritaet
		}
	}

	/** Empfaengerklasse fuer Archivdaten */
	public class ArchiveDataReceiver extends DataReceiver {

		public static final String MSG_PID_ATBACKSTEP_SUSPEND = "Archivzeit-Ruecksprung (suspend)";

		public static final String MSG_PID_ATBACKSTEP_RESUME = "Archivzeit-Ruecksprung (resume)";

		public static final String MSG_PID_ATFWSTEP_SUSPEND = "Archivzeit-Vorwaertssprung (suspend)";

		public static final String MSG_PID_ATFWSTEP_RESUME = "Archivzeit-Vorwaertssprung (resume)";

		private boolean suspendArchMessageSentBS = false, resumeArchMessageSentBS = false;	// Backstep

		private boolean suspendArchMessageSentFS = false, resumeArchMessageSentFS = false;	// Forward Step

		private boolean isSuspended = false;

		/**
		 * Wird auf false gesetzt, nachdem der erste Datensatz empfangen wurde. Wird verwendet, damit Vorwrtssprungs auch nach einem Neustart ses Archivsystems
		 * korrekt erkannt wird
		 */
		private boolean firstData = true;

		/** max. Archivzeitsprung vorwaerts (siehe TAnfArS 5.1.2.4.3.3) */
		private long tMaxATimeForwardStep = 3600 * 1000;	// Defaultwert 1 Stunde gemaess DatK

		private static final long STATISTIC_PRINT_TIME_INTERVAL = 60000;

		private long _statisticNextPrintTime;

		private long _statisticLastTotalReceivedDataSetCount = 0;

		private long _statisticLastTotalQueuedDataSetCount = 0;

		private long _statisticLastTotalSuccessedDataSetCount = 0;

		private long _statisticLastTotalFailedDataSetCount = 0;

		private long[] _statisticLastCacheCounts = {0, 0, 0, 0};


		public ArchiveDataReceiver() {
			logger.info("Ausgabe statistischer Informationen alle " + STATISTIC_PRINT_TIME_INTERVAL + " ms");
			_statisticNextPrintTime = (((ArchiveManager.archiveTime() + STATISTIC_PRINT_TIME_INTERVAL / 2 * 3) / STATISTIC_PRINT_TIME_INTERVAL)
			                           * STATISTIC_PRINT_TIME_INTERVAL);
		}

		public void processData(ResultData rd) {
			try {
				long aTime = ArchiveManager.archiveTime();

				if(aTime >= _statisticNextPrintTime) {
					StringBuilder message = new StringBuilder();
					final long totalReceivedDataSetCount = TOTAL_DS_RECEIVED.longValue();
					message.append("Aktivitt der Archiv-Queues im letzten Ausgabeintervall:");
					message.append(String.format("\n%20d Datenstze empfangen", totalReceivedDataSetCount - _statisticLastTotalReceivedDataSetCount));
					_statisticLastTotalReceivedDataSetCount = totalReceivedDataSetCount;

					final long totalQueuedDataSetCount = ArchiveTask.getTotalDS_Queued();
					message.append(String.format("\n%20d Datenstze weitergeleitet", totalQueuedDataSetCount - _statisticLastTotalQueuedDataSetCount));
					_statisticLastTotalQueuedDataSetCount = totalQueuedDataSetCount;

					final long totalSuccessedDataSetCount = ArchiveTask.getTotalDS_Success();
					message.append(
							String.format(
									"\n%20d Datenstze erfolgreich archiviert", totalSuccessedDataSetCount - _statisticLastTotalSuccessedDataSetCount
							)
					);
					_statisticLastTotalSuccessedDataSetCount = totalSuccessedDataSetCount;

					final long totalFailedDataSetCount = ArchiveTask.getTotalDS_Fail();
					message.append(
							String.format(
									"\n%20d Datenstze nicht erfolgreich archiviert", totalFailedDataSetCount - _statisticLastTotalFailedDataSetCount
							)
					);
					_statisticLastTotalFailedDataSetCount = totalFailedDataSetCount;

					message.append(String.format("\n%20d Datenstze noch in den Queues befindlich", onlineDataInQueues.intValue()));
					message.append(String.format("\n%19d%% Auslastung der Queues", (int)(getQueueLoad() * 100.0)));

					message.append("\nAktivitt der Zwischenspeicherung in geschlossenen Cache-Objekten: ");
					final long[] cacheCounts = CacheManager.getInstance().getCounts();
					message.append(String.format("\n%20d Datei-Schreibvorgnge", cacheCounts[0] - _statisticLastCacheCounts[0]));
					message.append(String.format("\n%20d verarbeitete Datenblcke", cacheCounts[1] - _statisticLastCacheCounts[1]));
					message.append(String.format("\n%20d zwischengespeicherte Datenblcke", cacheCounts[2] - _statisticLastCacheCounts[2]));
					message.append(String.format("\n%20d nicht zwischengespeicherte Datenblcke", cacheCounts[3] - _statisticLastCacheCounts[3]));
					_statisticLastCacheCounts = cacheCounts;

					long noDataIntervalCount = (aTime - _statisticNextPrintTime) / STATISTIC_PRINT_TIME_INTERVAL;
					_statisticNextPrintTime += STATISTIC_PRINT_TIME_INTERVAL * (noDataIntervalCount + 1);
					if(noDataIntervalCount > 0) message.append(String.format("\n%20d Intervalle ohne eingegangene Daten", noDataIntervalCount));
					logger.info(message.toString());
				}

				TOTAL_DS_RECEIVED.incrementAndGet();

				// Zu Testzwecken
				if(ArchiveManager.ONLY_DATA && !rd.hasData()) return;

				DataIdentNode din = didTree.getDataIdentNode(rd);
				if(din.arSParamIsArchivieren() && din.isArSParameterized()) {
					final long lastArchiveTime = ArchiveTask.getLastArchiveTime();
					// Folgende Abfrage sorgt dafr, dass kurze Rcksprnge der Zeit (bis 10 Sekunden) toleriert werden, in dem einfach die entsprechende
					// Zeit abgewartet wird und die aktuelle Zeit erneut abgefragt wird.
					if(aTime < lastArchiveTime) {
						long delta = lastArchiveTime - aTime;
						if(delta < 10000) {
							try { Thread.sleep(delta); } catch(InterruptedException ignored) { }
							aTime = ArchiveManager.archiveTime();
						}
					}
					if(!testForArchiveTimeBackstep(aTime, lastArchiveTime) && !testForArchiveTimeForwardStep(aTime, lastArchiveTime)) {
						putDataInQueue(aTime, rd);
						ArchiveTask.setLastArchiveTime(aTime);

						// Auf Ueberlast prfen und ggf. alle nicht schreibenden Tasks anhalten
						suspendNonWriteTasks();
					}
				}
				// else: kann nur vorkommen, wenn waehrend des Umparametrierens ein Datensatz eintrifft. Ist zu vernachlaessigen.
			}
			catch(DIdNodeNotFoundException e) {
				logger.error("Kein DataIdentNode fuer die zu archivierende Datenidentifikation vorhanden: " + Util.rd2Str(rd));
			}
		}

		public boolean mustSuspendQueries() {
			return isSuspended;
		}

		private boolean testForArchiveTimeForwardStep(long aTime, final long lastArchiveTime) {
			if(firstData) {	//nach einem Neustart des Archivsystems soll die berprfung erst beim zweiten Archivdatensatz vorgenommen werden
				firstData = false;
				return false;
			}
			else if(aTime - lastArchiveTime > tMaxATimeForwardStep) {
				if(!suspendArchMessageSentFS) {
					sendArchiveSuspendMsgFS(aTime);
					suspendArchMessageSentFS = true;
					resumeArchMessageSentFS = false;
				}
				return true;
			}
			else {
				if(suspendArchMessageSentFS && !resumeArchMessageSentFS) {
					sendArchiveResumeMsgFS(aTime);
					resumeArchMessageSentFS = true;
					suspendArchMessageSentFS = false;
				}
				return false;
			}
		}

		/**
		 * Prueft ob ein Ruecksprung der Archivzeit aufgetreten ist und sendet entsprechende Betriebsmeldungen.
		 *
		 * @param aTime aktuelle Archivzeit
		 *
		 * @param lastArchiveTime
		 * @return Wahr, wenn ein Ruecksprung aufgetreten ist, falsch sonst
		 */
		private boolean testForArchiveTimeBackstep(long aTime, final long lastArchiveTime) {
			if(aTime < lastArchiveTime) {
				if(!suspendArchMessageSentBS) {
					sendArchiveSuspendMsgBS(aTime);
					suspendArchMessageSentBS = true;
					resumeArchMessageSentBS = false;
				}
				return true;
			}
			else {
				if(suspendArchMessageSentBS && !resumeArchMessageSentBS) {
					sendArchiveResumeMsgBS(aTime);
					resumeArchMessageSentBS = true;
					suspendArchMessageSentBS = false;
				}
				return false;
			}
		}

		/**
		 * Legt das Datum in die passende Warteschlange. Falls die voll ist, wird einmal eine Log-Message ausgegeben. Wenn sie nicht mehr voll ist, wird ebenfalls
		 * einmal eine Log-Message ausgegeben.
		 *
		 * @param archiveTime Archivzeit des zu archivierenden Datums
		 * @param resultData  zu archivierendes Datum
		 */
		private void putDataInQueue(long archiveTime, ResultData resultData) {
			int tskIdx = calcOnlineArchivTaskIndex(resultData);

			boolean pushSuccessful = false;
			long nextTimeoutTime = archiveTime + SHIFT_QUEUES_INTERVAL;
			try {
				while(!(pushSuccessful = archiveQueues[tskIdx].push(resultData, archiveTime, 5000))) {
					nextTimeoutTime = managePushTimeout(nextTimeoutTime, "empfangener Online-Datensatz", tskIdx);
				}
			}
			catch(InterruptedException e) {
				// ignored
			}
			accountPushResult(pushSuccessful, tskIdx, "Archivierung", resultData);
		}

		private void sendArchiveSuspendMsgFS(long aTime) {
			String msg = "Es ist ein Vorwaertssprung der Archivzeit aufgetreten: die Archivzeit des soeben erhaltenen Datensatzes ("
			             + Util.timestrMillisFormatted(aTime) + ") liegt mehr als " + tMaxATimeForwardStep / 1000 + "sec ("
			             + Util.relTimestrMillis(tMaxATimeForwardStep) + ") vor der Archivzeit des zuletzt archivierten Datensatzes ("
			             + Util.timestrMillisFormatted(ArchiveTask.getLastArchiveTime())
			             + "). Das Archivsystem stellt die Archivierung ein, bis die Archivzeit wieder innerhalb der zulaessigen Spanne liegt.";
			MessageSender.getInstance().sendMessage(
					MSG_PID_ATFWSTEP_SUSPEND,
					de.bsvrz.sys.funclib.operatingMessage.MessageType.APPLICATION_DOMAIN,
					"",
					MessageGrade.FATAL,
					MessageState.MESSAGE,
					msg
			);
			logger.info("Betriebsmeldung:" + Debug.NEWLINE + msg);
		}

		private void sendArchiveResumeMsgFS(long aTime) {
			String msg = "Der Vorwaertssprung der Archivzeit ist behoben: die Archivzeit des soeben erhaltenen Datensatzes ("
			             + Util.timestrMillisFormatted(aTime) + ") liegt wieder innerhalb der zulaessigen Spanne von "
			             + tMaxATimeForwardStep / 1000 + "sec (" + Util.relTimestrMillis(tMaxATimeForwardStep)
			             + ") seit der Archivzeit des zuletzt archivierten Datensatzes (" + Util.timestrMillisFormatted(ArchiveTask.getLastArchiveTime())
			             + ". Das Archivsystem hat die Archivierung wieder aufgenommen.";
			MessageSender.getInstance().sendMessage(
					MSG_PID_ATFWSTEP_RESUME,
					de.bsvrz.sys.funclib.operatingMessage.MessageType.APPLICATION_DOMAIN,
					"",
					MessageGrade.INFORMATION,
					MessageState.MESSAGE,
					msg
			);
			logger.info("Betriebsmeldung:" + Debug.NEWLINE + msg);
		}

		private void sendArchiveSuspendMsgBS(long aTime) {
			String msg = "Es ist ein Ruecksprung der Archivzeit aufgetreten: die Archivzeit des zuletzt archivierten " + "Datensatzes ("
			             + Util.timestrMillisFormatted(ArchiveTask.getLastArchiveTime()) + ") liegt nach der Archivzeit des soeben erhaltenen Datensatzes ("
			             + Util.timestrMillisFormatted(aTime) + "). Das Archivsystem stellt die Archivierung ein, bis die Archivzeit wieder nach der Archivzeit"
			             + " des zuletzt archivierten Datensatzes liegt.";
			MessageSender.getInstance().sendMessage(
					MSG_PID_ATBACKSTEP_SUSPEND,
					de.bsvrz.sys.funclib.operatingMessage.MessageType.APPLICATION_DOMAIN,
					"",
					MessageGrade.FATAL,
					MessageState.MESSAGE,
					msg
			);
			logger.info("Betriebsmeldung:" + Debug.NEWLINE + msg);
		}

		private void sendArchiveResumeMsgBS(long aTime) {
			String msg = "Der Ruecksprung der Archivzeit ist behoben: die Archivzeit des zuletzt archivierten " + "Datensatzes ("
			             + Util.timestrMillisFormatted(ArchiveTask.getLastArchiveTime())
			             + ") liegt wieder vor der Archivzeit des soeben erhaltenen Datensatzes (" + Util.timestrMillisFormatted(aTime)
			             + "). Das Archivsystem hat die Archivierung wieder aufgenommen.";
			MessageSender.getInstance().sendMessage(
					MSG_PID_ATBACKSTEP_RESUME,
					de.bsvrz.sys.funclib.operatingMessage.MessageType.APPLICATION_DOMAIN,
					"",
					MessageGrade.INFORMATION,
					MessageState.MESSAGE,
					msg
			);
			logger.info("Betriebsmeldung:" + Debug.NEWLINE + msg);
		}

		/**
		 * Setzt den maximal erlaubten Vorwaertssprung der Archivzeit.
		 *
		 * @param tMax max. erlaubter Sprung (in Sekunden)
		 */
		public void setTMaxATimeForwardStep(long tMaxSeconds) {
			tMaxATimeForwardStep = tMaxSeconds * 1000;
		}

		public long getTMaxATimeForwardStep() {
			return tMaxATimeForwardStep;
		}
	}

	/**
	 * Abstrakte Klasse, die zum Empfang bestimmter Daten (z.B. von Archiv-Queries) verwendet werden kann. Auf diese Weise muss das eingehende Datum nicht mehr
	 * anhand von IDs unterschieden und der passenden Warteschlange zugewiesen werden.
	 */
	public abstract class DataReceiver implements ClientReceiverInterface {

		private String logPrefix = getClass().getSimpleName() + " update()";

		public void update(ResultData[] rds) {
			logger.finer(logPrefix, "ResultData[] " + (rds != null ? ("mit " + rds.length + " Datensaetzen") : "= null") + " erhalten");

			for(int i = 0; i < rds.length; i++) {
				try {

					// *** NUR FUER TESTZWECKE
					if(ArchiveManager.DATAINDEX_OFFSET != 0) {
						logger.warning("Datenindex-Offest aktiviert! (" + ArchiveManager.DATAINDEX_OFFSET + ")");
						try {
							JavaSpy.setFieldValue(rds[i], "dataIndex", rds[i].getDataIndex() + ArchiveManager.DATAINDEX_OFFSET);
						}
						catch(Exception e) {
							logger.error("Verstellen des Datenindex zu Testzwecken fehlgeschlagen: " + e.getMessage());
						}
					}
					// ***

					if(terminated) {
						logger.info("Archivsystem wird gerade beendet, Daten werden nicht mehr bearbeitet.");
						return;
					}
					// TAnfArS 5.1.2.4.3.7: initiale "keine Quelle"-DS ignorieren
					if(!isInitialNoSourceDS(rds[i])) {
						logger.finest(logPrefix + "ResultData[" + i + "]", rds[i]);
						processData(rds[i]);
					}
					else {
						logger.finest(logPrefix + "Initialer keine-Quelle-DS ignoriert", rds[i]);
					}
				}
				catch(Throwable t) {
					logger.warning("Fehler in der update()-Methode:\n" + Util.getStackTrace(t), rds[i]);
				}
			}
		}

		/**
		 * Prueft, ob es sich bei dem Datensatz um einen initialen "Keine Quelle"-Datensatz handelt, der vom DaV erzeugt wird, sobald sich ein Empfaenger angemeldet
		 * hat, ohne dass eine Quelle existiert.
		 *
		 * @param rd Datensatz
		 *
		 * @return Wahr, falls initialer "Keine Quelle"-Datensatz, falsch sonst
		 */
		private boolean isInitialNoSourceDS(ResultData rd) {
			return Util.dIdxSrcSubscrTime(rd) == 0 && Util.dIdxLfdnr(rd) == 0;
		}

		public abstract void processData(ResultData rd);
	}

	public class DataAckSender implements ClientSenderInterface {

		public void sendAck(ResultData rd, long qAspID) throws ConfigurationException, DataNotSubscribedException, SendSubscriptionNotConfirmed {
			ResultData qRd = new ResultData(
					rd.getObject(),
					new DataDescription(rd.getDataDescription().getAttributeGroup(), archMgr.getAsp(qAspID)),
					ArchiveManager.archiveTime(),
					rd.getData()
			);
			archMgr.getDavCon().sendData(new ResultData[]{qRd});
		}

		/** @see ClientSenderInterface#isRequestSupported(SystemObject,DataDescription) */
		public boolean isRequestSupported(SystemObject object, DataDescription dataDescription) {
			return true;
		}

		/** @see ClientSenderInterface#dataRequest(SystemObject,DataDescription,byte) */
		public void dataRequest(SystemObject object, DataDescription dataDescription, byte state) {
			if(state == ClientSenderInterface.STOP_SENDING_NOT_A_VALID_SUBSCRIPTION) {
				logger.warning(
						"Sendesteuerung fuer Quittung '" + object + ", " + dataDescription + "': STOP_SENDING_NOT_A_VALID_SUBSCRIPTION" + Debug.NEWLINE
						+ "Moeglicherweise ist unter dieser Datenidentifikation bereits eine Quelle angemeldet."
				);
			}
			else if(state == ClientSenderInterface.STOP_SENDING_NO_RIGHTS) {
				logger.warning("Sendesteuerung fuer Quittung " + object + ", " + dataDescription + " : STOP_SENDING_NO_RIGHTS");
			}
		}
	}
}
