/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 * 
 * This file is part of de.bsvrz.ars.ars.
 * 
 * de.bsvrz.ars.ars is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.ars.ars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.ars.ars; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

package de.bsvrz.ars.ars.mgmt.simulation;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.ClientReceiverInterface;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.DataState;
import de.bsvrz.dav.daf.main.ReceiveOptions;
import de.bsvrz.dav.daf.main.ReceiverRole;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.SenderRole;
import de.bsvrz.dav.daf.main.ClientSenderInterface;
import de.bsvrz.dav.daf.main.OneSubscriptionPerSendData;
import de.bsvrz.dav.daf.main.SendSubscriptionNotConfirmed;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.ConfigurationException;
import de.bsvrz.dav.daf.main.config.DataModel;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.losb.datk.PidSimulation;
import de.bsvrz.sys.funclib.losb.exceptions.FailureException;
import de.bsvrz.sys.funclib.losb.exceptions.SenderException;
import de.bsvrz.sys.funclib.losb.kernsoftware.ConnectionManager;
import de.bsvrz.sys.funclib.losb.kernsoftware.SimpleSender;
import de.bsvrz.sys.funclib.losb.messages.ErrorMessage;

import java.util.*;

/**
 * Deterministischer endlicher Automat fr die Simulation. Nicht erlaubte Transitionen werden ignoriert. Treten Fehler auf, so werden diese geloggt.
 *
 * @author beck et al. projects GmbH
 * @author Martin Hilgers
 * @version $Revision: 6603 $ / $Date: 2009-04-27 10:50:08 +0200 (Mo, 27 Apr 2009) $ / ($Author: rs $)
 */
public class SimulationAutomaton implements ClientReceiverInterface {
	/////////////////////////////
	/// Zustand des Automaten ///
	/////////////////////////////

	/** Undefinierter Zustand / Undefinierte Simulationsvariante */
	public static final int UNDEFINED = -1;

	/** Automat wurde erzeugt, ist aber noch nicht aktiv. */
	private static final int CREATED = -2;

	/** Automat wurde erzeugt und wurde aktiviert - d.h. Anmeldung auf Simulationssteuerung ist erfolgt. */
	private static final int SUBSCRIBED = -3;

	/*
		 * !!!WICHTIG!!!
		 * Die folgenden Werte mssen mit denen aus dem Datenkatalog bereinstimmen!
		 */
	/** Neu - Zustand. */
	protected static final int NEW = 0;

	/** Vorstart - Zustand. */
	protected static final int INIT = 1;

	/** Start - Zustand. */
	protected static final int START = 2;

	/** Pause - Zustand. */
	protected static final int PAUSE = 5;

	/** Stop - Zustand. */
	protected static final int STOP = 3;

	/** Gelscht - Zustand. */
	protected static final int ERASED = 4;

	/** Zustand des Automaten. */
	private int state = CREATED;

	/////////////////////////////
	/// Eigenschaften des DFA ///
	/////////////////////////////	
	/** Debug Ausgabe. */
	private Debug debug = Debug.getLogger();

	/** Simulationsobjekt */
	private SystemObject simulation;

	/** Simulationsstreckenobjekt. */
	private SystemObject simulationRange = null;

	/** Liste mit den Transitionen, die ausgefhrt werden sollen. */
	private ArrayList<Integer> transitions;

	/** Simulationsvariante der Simulation. */
	private short simVar = UNDEFINED;

	/** Zeigt an, dass der letzte erhaltene Datensatz fr das SImulationsobjekt ein 'keine Quelle' Datensatz war. */
	private boolean noSource = false;

	/** Steuerungsinformationen, auf die sich die Simulation beim Dav anmelden soll. */
	private final DataDescription simulationControl;

	/** Simulationsverwaltung */
	private final SimulationManager simManager;

	private StateSender _stateSender;


	/**
	 * Erzeugt einen neuen Automaten fr eine Simulation. Meldet den Automaten als Empfnger fr das Simuationsobjekt an.
	 *
	 * @param simManager        Simulationsverwaltung
	 * @param simulation        Simulationsobjekt.
	 * @param simulationControl Steuerungsinformationen, auf die sich die Simulation beim Dav anmelden soll.
	 */
	public SimulationAutomaton(SimulationManager simManager, SystemObject simulation, DataDescription simulationControl) {
		if(simulationControl.getAspect() == null || simulationControl.getAttributeGroup() == null) {
			throw new IllegalArgumentException("Ungltige DataDescription: " + simulationControl);
		}
		else if(simulation == null) throw new IllegalArgumentException("Simulationsobjekt ist null.");
		this.simManager = simManager;
		this.simulation = simulation;
		this.simulationControl = simulationControl;
		transitions = new ArrayList<Integer>();

		//Daten der Simulation abrufen
		AttributeGroup atgProperties = simManager.getDav().getDataModel().getAttributeGroup(PidSimulation.atgProperties);
		Data properties = simulation.getConfigurationData(atgProperties);
		this.simulationRange = properties.getReferenceValue(PidSimulation.attSimulationRange).getSystemObject();
		this.simVar = properties.getUnscaledValue(PidSimulation.attSimVar).shortValue();
		if(simVar < 0 || simVar > 999) throw new IllegalStateException("Simulationsvariante auerhalb des Wertebereiches: " + properties);
	}


	/** Meldet das Objekt als Empfnger fr die Simulationssteuerung an. */
	private void subscribeSimulation() {
		try {
			ConnectionManager.subscribeReceiver(simManager.getDav(), this, simulation, simulationControl, ReceiveOptions.normal(), ReceiverRole.receiver());
		}
		catch(ConfigurationException e) {
			debug.error(ErrorMessage.COMMUNICATION, e);
		}
		catch(FailureException e) {
			debug.error("", e);
		}
	}


	/** Meldet das Objekt vom Empfang fr Simulationssteuerungsdaten ab. */
	private void unsubscribeSimulation() {
		ConnectionManager.unsubscribeReceiver(simManager.getDav(), this, simulation, simulationControl);
	}


	/** Meldet das Objekt als Empfnger fr die Simulationsstrecke an. */
	private void subscribeSimulationRange() {
		try {
			DataDescription dd = new DataDescription(
					simManager.getDav().getDataModel().getAttributeGroup(PidSimulation.atgArchiveData),
					simManager.getDav().getDataModel().getAspect(PidSimulation.aspTarget)
			);
			ConnectionManager.subscribeReceiver(simManager.getDav(), this, simulationRange, dd, ReceiveOptions.normal(), ReceiverRole.receiver());
		}
		catch(ConfigurationException e) {
			debug.error(ErrorMessage.COMMUNICATION, e);
		}
		catch(FailureException e) {
			debug.error("", e);
		}
	}


	/** Meldet das Objekt vom Empfang fr Simulationsstreckendaten ab. */
	private void unsubscribeSimulationRange() {
		DataDescription dd = new DataDescription(
				simManager.getDav().getDataModel().getAttributeGroup(PidSimulation.atgArchiveData),
				simManager.getDav().getDataModel().getAspect(PidSimulation.aspTarget)
		);
		ConnectionManager.unsubscribeReceiver(simManager.getDav(), this, simulationRange, dd);
	}


	/**
	 * Transition in den neuen Zustand. Es wird berprft, ob die Transition mglich ist.
	 *
	 * @return <code>true</code> falls der Automat in den neuen Zustand berfhrt werden konnte. <code>false</code> sonst.
	 */
	public synchronized boolean transition() {
		if(state == CREATED)			//der Erste Aufruf von Transition fhrt zur Initialisierung
		{
			state = SUBSCRIBED;
			subscribeSimulation();
			subscribeSimulationState();
			return true;
		}
		else if(transitions.isEmpty()) return false;

		int newState = transitions.remove(0);
		/*
		 * Erlaubte Transitionen
		 * UNDEFINED -> NEW
		 * NEW -> INIT
		 * INIT -> START
		 * START -> PAUSE | STOP
		 * PAUSE -> PAUSE | START
		 * STOP -> INIT | ERASED 
		 */
		if((state == SUBSCRIBED && newState == NEW) ||		//Nur gltige Transitionen ausfhren
		   (state == NEW && newState == INIT) || (state == INIT && newState == START) || (state == START && (newState == STOP || newState == PAUSE))
		   || (state == PAUSE && (newState == START || newState == PAUSE || newState == STOP)) || (state == STOP && (newState == INIT || newState == ERASED))) {
			debug.finest("Zustandsbergang: " + state + " -> " + newState);
			state = (int)newState;
			switch(state) {
				case NEW:
					enterNewState();
					break;
				case INIT:
					enterInitState();
					break;
				case START:
					enterStartState();
					break;
				case STOP:
					enterStopState();
					break;
				case ERASED:
					enterErasedState();
			}
			return true;
		}
		else if(state == SUBSCRIBED) {
			if(newState
			   == INIT)		//Archivsystem wurde erst hochgefahren, nachdem die Simulation initialisiert wurde -> Simulation kann trotzdem durchgefhrt werden
			{
				enterNewState();
				state = INIT;
				enterInitState();
				return true;
			}
			else						//Simulation ist bereits zu weit fortgeschritten
			{
				sendSimulationState(false);
				return false;
			}
		}
		else							//alle brigen Zustandsbergnge werden ignoriert
		{
			debug.fine("Nicht erlauber Zustandsbergang: " + state + " -> " + newState);
			return false;
		}
	}


	/**
	 * Zustand 'Neu' wird erreicht. Meldet den Empfang von Steuerungsdaten an. Sendet, dass das Archivsystem nicht fr die Durchfhrung der Simulation zur
	 * Verfgung steht.
	 */
	protected void enterNewState() {
		sendSimulationState(false);
	}

	/*
		 * Ablauf fr die Methoden enterInitState, continueInitStateGetParams, continueInitStateSubscribe & finishInitState:
		 * 1. Aufruf von enterInitState 											(Wirf Lschtask an)
		 * 2. Lschtask Callback -> dataDeleted ruft continueInitStateGetParams auf
		 * 3. continueInitStateGetParams											(meldet sich auf Dav Daten an um zu erfahren, welche Daten zu archivieren sind)
		 *  -> Dav Callback -> update()
		 *  4. update ruft continueInitStateSubscribe() auf. 						(Dort werden die Simulationsdaten, auf die sich das Archivsystem anmelden soll, ausgewertet)
		 *  5. ArchivConfig Callback -> finishInitState 							(Das Archivsystem hat sich auf die Simulationsdaten angemeldet.)
		 */

	/**
	 * Zustand 'Vorstart' wird erreicht. - Lschen der archivierten Simulationsdaten. Ist diese Aufgabe abgeschlossen, wird die Callback Methode{@link
	 * #dataDeleted(boolean)} vom Lschtask aufgerufen.
	 */
	protected void enterInitState() {
		//Lschen der Simulationsdaten
		SimulationResultData ard = new SimulationResultData(this);
		simManager.insertDeleteQuery(ard);
	}


	/**
	 * - Anmelden auf Information, welche Daten archiviert werden sollen. Geht diese Information ein, wird {@link #continueInitStateSubscribe(ResultData)} durch
	 * {@link #update(ResultData[])} aufgerufen. Diese Methode Wird von {@link #dataDeleted(boolean)} aufgerufen, wenn der Zustand der Simulation {@link #INIT}
	 * ist.
	 */
	protected void continueInitStateGetParams() {
		//Daten herausfinden, die Archiviert werden sollen.
		subscribeSimulationRange();
	}


	/**
	 * Fortsetzung Zustand 'Vorstart'. - Meldet das Archivsystem auf die Simulationsdaten an. Wird durch {@link #update(ResultData[])} aufgerufen, sobald die zu
	 * archivierenden Daten bekannt sind.
	 *
	 * @param resultData Information, auf welche Daten sich das Archivsystem anmelden soll.
	 */
	protected void continueInitStateSubscribe(ResultData resultData) {
		SimulationResultData ard = new SimulationResultData(this, resultData);
		if(!simManager.getArchMgr().getInQueuesMgr().insertSimVarParam(ard)) {
			debug.warning("Parametrierungsqueue ist voll. Die Simulation kann nicht durchgefhrt werden.");
			terminate();
		}
	}


	/**
	 * Wird vom ArchivConfig Task aufgerufen, nachdem das Archivsystem sich auf die Simulationsdaten angemeldet hat - Sendet die 'Bereit' Meldung, wenn Daten
	 * gelscht, falls der Automat im Zustand Vorstart und die Anmeldung auf die zu archivierenden Simulationsdaten erfolgreich verlaufen ist.
	 *
	 * @param successful Zeigt an ob der Auftrag erfolgreich durchgefhrt wurde.
	 */
	protected void finishInitState(boolean successful) {
		if(!successful) state = UNDEFINED;
		sendSimulationState(successful);
	}


	/** Zustand 'Start' wird erreicht. Meldet die Simulation vom Empfang von Anmeldeinformationen */
	protected void enterStartState() {
		unsubscribeSimulationRange();
	}


	/** Zustand 'Stop' wird erreicht. - 'Ende'-Meldung senden - Abmeldung vom Datenempfang */
	protected void enterStopState() {
		sendSimulationState(false);
		simManager.getArchMgr().getArchivConfig().unsubsrcibeSimVarData(this.getSimVar());
	}


	/** Zustand 'Gelscht' wird erreicht. - Lschen der Daten */
	protected void enterErasedState() {
		SimulationResultData srd = new SimulationResultData(this);
		simManager.insertDeleteQuery(srd);
	}

	/**
	 * Wird aufgerufen, nachdem alle Daten gelscht wurden. Wird durch {@link #dataDeleted(boolean)} aufgerufen, wenn der Zustand {@link #ERASED} ist. -
	 * Benachrichtigung senden, dass Simulationsobjekt gelscht werden kann.
	 */
	protected void finishErasedState() {
		sendSimulationState(false);
		unsubscribeSimulationState();
	}


	private void subscribeSimulationState() {
		final ClientDavInterface daf = simManager.getDav();
		DataModel model = daf.getDataModel();
		AttributeGroup atg = model.getAttributeGroup(PidSimulation.atgArchiveState);
		DataDescription dd = new DataDescription(atg, model.getAspect(PidSimulation.aspState));
		try {
			_stateSender = new StateSender();
			daf.subscribeSender(_stateSender, simulation, dd, SenderRole.source());
		}
		catch(OneSubscriptionPerSendData e) {
			debug.warning("Mehrfachanmeldung auf Simulationszustand, Simulation: " + simulation, e);
		}
	}

	private void unsubscribeSimulationState() {
		final StateSender stateSender = _stateSender;
		if(stateSender != null) {
			_stateSender = null;
			final ClientDavInterface daf = simManager.getDav();
			DataModel model = daf.getDataModel();
			AttributeGroup atg = model.getAttributeGroup(PidSimulation.atgArchiveState);
			DataDescription dd = new DataDescription(atg, model.getAspect(PidSimulation.aspState));
			try {
				daf.unsubscribeSender( stateSender, simulation, dd);
			}
			catch(Exception e) {
				debug.warning("Abmeldung auf Simulationszustand nicht mglich, Simulation: " + simulation, e);
			}
		}
	}

	private static class StateSender implements ClientSenderInterface {

		public void dataRequest(SystemObject object, DataDescription dataDescription, byte state) {
		}

		public boolean isRequestSupported(SystemObject object, DataDescription dataDescription) {
			return false;
		}
	}

	/**
	 * Legt den Simulationsstatus auf den Datenverteiler. Blockiert, bis das Senden abgeschlossen wurde, maximal jedoch 10 Sekunden.
	 *
	 * @param archiveReady Ist das Archiv fr die Simulation bereit?
	 *
	 * @return <code>true</code> falls das Senden erfolgreich durchgefhrt wurde.
	 */
	protected boolean sendSimulationState(boolean archiveReady) {
		if(_stateSender == null) {
			debug.warning("Versand des Simulationszustands nicht mglich, weil nicht angemeldet, Simulation: " + simulation);
			return false;
		}
		final ClientDavInterface daf = simManager.getDav();
		DataModel model = daf.getDataModel();
		AttributeGroup atg = model.getAttributeGroup(PidSimulation.atgArchiveState);
		DataDescription dd = new DataDescription(atg, model.getAspect(PidSimulation.aspState));
		Data data = daf.createData(atg);

		data.getUnscaledValue(PidSimulation.attArchiveReady).set((archiveReady) ? 1 : 0);
		try {
			daf.sendData(new ResultData(simulation, dd, System.currentTimeMillis(), data));
			return true;
		}
		catch(SendSubscriptionNotConfirmed e) {
			debug.warning("Versand des Simulationszustands nicht mglich, Simulation: " + simulation, e);
			return false;
		}
	}

	/**
	 * Legt den Simulationsstatus auf den Datenverteiler. Wartet nicht.
	 *
	 * @param archiveReady Ist das Archiv fr die Simulation bereit?
	 */
	protected void sendSimulationStateAndUnsubscribe(boolean archiveReady) {
		if(_stateSender != null) {
			sendSimulationState(archiveReady);
			unsubscribeSimulationState();
		}
	}


	/**
	 * Wird aufgerufen, nachdem die Daten der Simulation gelscht wurden.
	 *
	 * @param successful Zeigt an ob die Daten gelscht werden konnten.
	 */
	public void dataDeleted(boolean successful) {
		if(!successful) debug.warning("Es konnten nicht alle Simulationsdaten gelscht werden.");
		if(state == INIT) {
			continueInitStateGetParams();
		}
		else if(state == ERASED) finishErasedState();
	}


	/** Beendet den Automaten. Meldet den Empfang von Daten ab. */
	public void terminate() {
		unsubscribeSimulation();
		//ggf Meldung absetzten
		if(state != ERASED) sendSimulationStateAndUnsubscribe(false);
		state = UNDEFINED;
	}


	/**
	 * Callback Methode die vom Dav aufgerufen wird, wenn neue Daten zum Simulationsobjekt eingehen. Benachrichtigt Simulationsverwaltung, wenn eine
	 * Zustandsnderung der Simulation eingegangen ist.
	 *
	 * @see ClientReceiverInterface#update(ResultData[])
	 * @see SimulationManager#addTransition(SimulationAutomaton)
	 */
	public void update(ResultData[] results) {
		for(ResultData result : results) {
			try {
				//berprfen, zu welchem Objekt die Daten gehren
				if(result.getObject().equals(simulation)) {
					if(result.getDataState().equals(DataState.NO_SOURCE)) {
						if(state != SUBSCRIBED) {
							//'keine Quelle' Datensatz -> auf weiteren Steuerbefehl warten
							noSource = true;
							simManager.addWakeUp(this);
						}
						else {
							//Archivsystem ist hochgefahren und hat nur den keine Quelle Datensatz zur Verfgung: 
							//Meldung Senden, dass das Archivsystem nicht fr die Simulation nicht zur Verfgung steht!
							sendSimulationState(false);
						}
					}
					else if(result.hasData()) {
						//Es handelt sich um eine Zustandsnderung
						if(noSource) {
							noSource = false;
							simManager.cancelWakeUp(this);
						}
						Data data = result.getData();
						Integer i = new Integer((int)data.getUnscaledValue(PidSimulation.attSimulationState).getState().getValue());
						transitions.add(i);
						simManager.addTransition(this);
					}
				}
				else if(result.getObject().equals(simulationRange) && result.hasData()) {
					//Es handelt sich um Informationen zur Parametrierung
					continueInitStateSubscribe(result);
				}
			}
			catch(NullPointerException npe) {
				Debug.getLogger().warning("Fehler beim Auswerten der eingehenden Daten. Daten werden ignoriert! " + result, npe);
			}
		}
	}


	/**
	 * Benachrichtigt die Simulation, dass die Wartezeit abgelaufen ist.
	 *
	 * @see SimulationManager#addWakeUp(SimulationAutomaton)
	 */
	public void wakeUp() {
		if(!sendSimulationState(false)) debug.fine("Kontakt zur Smulationssteuerung verloren. Senden einer Statusmeldung nicht mglich.");
		simManager.getArchMgr().getArchivConfig().unsubsrcibeSimVarData(this.getSimVar());
	}


	/**
	 * Liefert die Simulationsvariante der Simulation.
	 *
	 * @return Simulationsvariante. {@link #UNDEFINED} falls noch nicht initialisiert.
	 */
	public short getSimVar() {
		return simVar;
	}
}
