/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 * 
 * This file is part of de.bsvrz.ars.ars.
 * 
 * de.bsvrz.ars.ars is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.ars.ars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.ars.ars; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

package de.bsvrz.ars.ars.mgmt.tasks;

import de.bsvrz.ars.ars.mgmt.ArchiveManager;
import de.bsvrz.ars.ars.mgmt.datatree.DIdNodeNotFoundException;
import de.bsvrz.ars.ars.mgmt.tasks.tasklistener.RingOfDeathListener;
import de.bsvrz.ars.ars.persistence.PersistenceException;
import de.bsvrz.ars.ars.persistence.index.IndexException;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.archive.ArchiveData;
import de.bsvrz.sys.funclib.losb.ringbuffer.ArchiveRingBuffer;
import de.bsvrz.sys.funclib.losb.util.Util;

import java.util.concurrent.atomic.AtomicLong;

/**
 * Task zur Archivierung nachgefordeter Daten. Unterscheidet sich hauptsaechlich in der Behandlung von Datenluecken von {@link ArchiveTask}.
 *
 * @author beck et al. projects GmbH
 * @author Alexander Schmidt
 * @version $Revision: 6419 $ / $Date: 2009-03-10 23:15:57 +0100 (Di, 10 Mrz 2009) $ / ($Author: rs $)
 */
public class ArchiveRequestedTask extends ArchiveTask {


	private static AtomicLong TOTAL_REQ_DS_QUEUED = new AtomicLong(0);

	private static AtomicLong TOTAL_REQ_DS_SUCCESS = new AtomicLong(0);

	private static AtomicLong TOTAL_REQ_DS_FAIL = new AtomicLong(0);


	private ArchiveData ad;

	/**
	 * Erzeugt einen ArchiveRequestedTask
	 *
	 * @param archiveMgr Archiv-Manager
	 * @param rodl       Ring-of-Death-Listener
	 * @param tskIndex   Index des Tasks
	 * @param inQueue    Eingangs-Warteschlange
	 */
	public ArchiveRequestedTask(ArchiveManager archiveMgr, RingOfDeathListener rodl, int tskIndex, ArchiveRingBuffer inQueue) {
		super(archiveMgr, rodl, tskIndex, inQueue);
	}

	/**
	 * Alle Instanzvariablen vorbelegen.
	 *
	 * @param archiveData Archivdatensatz
	 * @param archTime    Archivzeitstempel
	 *
	 * @throws DIdNodeNotFoundException
	 */
	private void presetData(ArchiveData archiveData, long archTime) throws DIdNodeNotFoundException {
		ad = archiveData;
		DataDescription dd = ad.getDataDescription();
		objID = ad.getObject().getId();
		atg = dd.getAttributeGroup();
		aspID = dd.getAspect().getId();
		sv = 0;
		arsTime = archTime;
		dataTime = ad.getDataTime();
		dataIdx = ad.getDataIndex();
		adk = ad.getDataKind().isDelayed() ? NN : NA;
		didPath = persMgr.getPathName(objID, atg.getId(), aspID, 0, adk);
		din = didTree.getDataIdentNode(objID, atg.getId(), aspID, 0);
	}

	public void work(Object archiveData, long archTime) {
		logReceivedData((ArchiveData)archiveData);

		// Synchronisierung ueber den DataIdentNode:
		boolean criticalAreaEntered = false;
		try {
			// Flags, IDs, Variable besetzen, DataIdentNode holen, etc.:
			presetData((ArchiveData)archiveData, archTime);

			criticalAreaEntered = din.enterCriticalArea();
			if(!criticalAreaEntered) {
				logger.warning("Archivierung nachgeforderter Daten abgebrochen: " + ad);
				return;
			}

			assignIndexes(REQUESTED);			// Indexe zuweisen, falls nicht schon geschehen
			lookupLastDataIdxes(REQUESTED);		// Letzte Datenindexe, Datenzeiten nachsehen	

			openContID = din.getOpenContID(adk);
			dataEntryLen = persMgr.serializeReqData(this, ad, arsTime, compresser);
			(storeData() ? TOTAL_REQ_DS_SUCCESS : TOTAL_REQ_DS_FAIL).incrementAndGet();
		}
		catch(DIdNodeNotFoundException e) {
			handleErr(e, Util.ad2Str(ad));
		}	// Darf nicht sein, weil die DID zum Archivieren parametriert wurde
		catch(PersistenceException e) {
			handleErr(e, Util.ad2Str(ad));
		}
		catch(IndexException e) {
			handleErr(e, Util.ad2Str(ad));
		}
		catch(Throwable e) {
			handleSevereErr(e, Util.ad2Str(ad));
		}

		finally {
			if(din != null && criticalAreaEntered) din.exitCriticalArea();
		}
	}

	protected void logReceivedData(ArchiveData ad) {
		logger.finer(getName() + ": nachgeforderten Datensatz zur Archivierung erhalten: ", Util.ad2Str(ad));
	}


	/**
	 * Wird vom {@link de.bsvrz.ars.ars.mgmt.InQueuesMgr.ArchiveDataReceiver} aufgerufen, nachdem er einen zu archivierenden Datensatz in eine der Queues gelegt
	 * hat.
	 */
	public static void incrementTotalDS_Queued() {
		TOTAL_REQ_DS_QUEUED.incrementAndGet();
	}

	/** @return Anzahl der seit Systemstart in die Warteschlangen eingefuegten zu archivierenden nachgeforderten Datensatze */
	public static long getTotalDS_Queued() {
		return TOTAL_REQ_DS_QUEUED.longValue();
	}

	/** @return Anzahl der seit Systemstart zur Archivierung verarbeiteten Datesaetze (erfolgreich und nicht erfolgreich archiviert) */
	public static long getTotalDS_Processed() {
		return TOTAL_REQ_DS_SUCCESS.longValue() + TOTAL_REQ_DS_FAIL.longValue();
	}

	/** @return Anzahl der seit Systemstart erfolgreich archivierten Datesaetze */
	public static long getTotalDS_Success() {
		return TOTAL_REQ_DS_SUCCESS.longValue();
	}

	/** @return Anzahl der seit Systemstart nicht erfolgreich archivierten Datesaetze */
	public static long getTotalDS_Fail() {
		return TOTAL_REQ_DS_FAIL.longValue();
	}

	/** Setzt alle Zaehler auf 0. */
	public static void resetDSCounter() {
		TOTAL_REQ_DS_SUCCESS.set(0);
		TOTAL_REQ_DS_FAIL.set(0);
		TOTAL_REQ_DS_QUEUED.set(0);
	}

}
