/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 * 
 * This file is part of de.bsvrz.ars.ars.
 * 
 * de.bsvrz.ars.ars is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.ars.ars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.ars.ars; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

package de.bsvrz.ars.ars.mgmt.tasks.scheduler;

import de.bsvrz.ars.ars.mgmt.ArchiveManager;
import de.bsvrz.ars.ars.mgmt.tasks.BackupTask;
import de.bsvrz.ars.ars.mgmt.tasks.LZVTask;
import de.bsvrz.ars.ars.mgmt.tasks.RequestGapTask;
import de.bsvrz.sys.funclib.debug.Debug;

import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.*;

/**
 * Diese Klasse verwaltet einen Timer und prft im Minutentakt, ob die Tasks fr Sicherung, Lschen und Nachforndern ausgefhrt werden mssen.
 * <p/>
 * ber die public-Methoden knnen die IntervalWerte gesetzt werden.
 *
 * @author beck et al. projects GmbH
 * @author Phil Schrettenbrunner
 * @version $Revision: 10735 $ / $Date: 2012-11-19 18:14:45 +0100 (Mo, 19 Nov 2012) $ / ($Author: rs $)
 */
public class TaskScheduler {

	public static final DateFormat dateFormat = new SimpleDateFormat("dd.MM.yyyy HH:mm:ss");

	private static final int ONE_MINUTE = 60 * 1000;

	private static Debug logger;

	private boolean terminated = false;

	private Timer timer;

	private long lzvInterval;

	private ArchiveManager archiveManager;

	private TimeIntervalCron backupTimeInterval = new TimeIntervalCron();

	private TimeIntervalCron deleteTimeInterval = new TimeIntervalCron();

	private TimeIntervalCron requestTimeInterval = new TimeIntervalCron();

	private Date backupLastRun = null;

	private Date deleteLastRun = null;

	private Date requestLastRun = null;

	private Date lzvLastRun = null;

	private CronTask backupCronTask;

	private CronTask deleteCronTask;

	private CronTask requestCronTask;

	private IntervalTask lzvTask;


	/**
	 * Erzeugt einen TaskScheduler mit den angegebenen Interval-Zeiten fr den Sicherungs-, Lsch- und Nachforderungstask.
	 *
	 * @param davConnection Dav Connection
	 */
	public TaskScheduler(ArchiveManager archiveManager) {
		logger = Debug.getLogger();
		this.timer = new Timer();
		this.archiveManager = archiveManager;
		backupCronTask = new CronTask(backupTimeInterval);
		deleteCronTask = new CronTask(deleteTimeInterval);
		requestCronTask = new CronTask(requestTimeInterval);
	}


	/** Startet den Timer, der die Tasks steuert */
	public void start() {
		// Timer zur nchsten vollen Minute starten
		long start = ((System.currentTimeMillis() / 60000) * 60000 + 60000) - System.currentTimeMillis();
		start = start < 100 ? 0 : start;
		timer.scheduleAtFixedRate(backupCronTask, start, ONE_MINUTE);
		timer.scheduleAtFixedRate(deleteCronTask, start, ONE_MINUTE);
		timer.scheduleAtFixedRate(requestCronTask, start, ONE_MINUTE);
	}


	public void setIntervals(String backup, String delete, String request, long lzv) {
		backupTimeInterval.setFields(backup);
		if(backupTimeInterval.hasParseErrors()) {
			logger.warning(
					"'" + backup + "' ist kein gltiger ZeitIntervallCron fr das Sicherungsintervall. Automatischer Start fr die Sicherung DEAKTIVIERT"
			);
		}
		else {
			logger.fine("Setze Sicherungsintervall auf: " + backupTimeInterval.toString());
		}

		deleteTimeInterval.setFields(delete);
		if(deleteTimeInterval.hasParseErrors()) {
			logger.warning("'" + delete + "' ist kein gltiger ZeitIntervallCron fr das Lschintervall. Automatischer Start fr das Lschen DEAKTIVIERT");
		}
		else {
			logger.fine("Setze Lschintervall auf: " + deleteTimeInterval.toString());
		}

		requestTimeInterval.setFields(request);
		if(requestTimeInterval.hasParseErrors()) {
			logger.warning(
					"'" + request + "' ist kein gltiger ZeitIntervallCron fr das Nachforderungsintervall. Automatischer Start fr das Nachfordern DEAKTIVIERT"
			);
		}
		else {
			logger.fine("Setze Nachforderungsintervall auf: " + requestTimeInterval.toString());
		}

		if(lzvTask != null) lzvTask.cancel(); // beendet ev. vorhandenen Timer
		if(lzv < ONE_MINUTE) lzv = ONE_MINUTE; // 1 Min ist kleinster Wert
		lzvTask = new IntervalTask();
		if(!terminated) timer.scheduleAtFixedRate(lzvTask, lzv, lzv);
		lzvInterval = lzv;
	}


	/**
	 * Das Interval als String. Wenn leerer String "", wird der Task nie ausgefhrt.
	 *
	 * @return Das Interval (ausgewertet) als String
	 */
	public String getBackupIntervals() {
		return backupTimeInterval.toString();
	}


	/**
	 * Liefert das Datum, an dem der Task das letzte mal gestartet wurde. Oder einen leeren String, wenn der Task noch nie gestartet wurde.
	 *
	 * @return Datum als String
	 */
	public String getBackupLastRun() {
		if(backupLastRun != null) return dateFormat.format(backupLastRun);
		return "(wurde noch nie gestartet)";
	}


	/**
	 * Liefert das Datum, an dem der Task das nchste mal gestartet wird. Oder einen String, der besagt, dass der Task nie gestartet wird (weil die
	 * konfigurierten Werte ungltig sind). Bercksichtigt werden Daten, die maximal 1 Jahr und 1 Tag in der Zukunft liegen. Tritt kein Fehler auf und wird
	 * trotzdem kein Datum gefunden ist der Rckgabewert "> 1 Jahr".
	 *
	 * @return Datum als String
	 */
	public String getBackupNextRun() {
		if(backupTimeInterval.hasValues()) {
			Calendar backupNextRun = backupTimeInterval.getNextRun(null); // Parameter=null bedeutet "von jetzt ab"
			if(backupNextRun != null) {
				return dateFormat.format(backupNextRun.getTime());
			}
			else {
				return "> 1 Jahr";
			}
		}
		return "---";
	}


	/**
	 * Das Interval als String. Wenn leerer String "", wird der Task nie ausgefhrt.
	 *
	 * @return Das Interval (ausgewertet) als String
	 */
	public String getDeleteIntervals() {
		return deleteTimeInterval.toString();
	}


	/**
	 * Liefert das Datum, an dem der Task das letzte mal gestartet wurde. Oder einen leeren String, wenn der Task noch nie gestartet wurde.
	 *
	 * @return Datum als String
	 */
	public String getDeleteLastRun() {
		if(deleteLastRun != null) return dateFormat.format(deleteLastRun);
		return "(wurde noch nie gestartet)";
	}


	/**
	 * Liefert das Datum, an dem der Task das nchste mal gestartet wird. Oder einen String, der besagt, dass der Task nie gestartet wird (weil die
	 * konfigurierten Werte ungltig sind). Bercksichtigt werden Daten, die maximal 1 Jahr und 1 Tag in der Zukunft liegen. Tritt kein Fehler auf und wird
	 * trotzdem kein Datum gefunden ist der Rckgabewert "> 1 Jahr".
	 *
	 * @return Datum als String
	 */
	public String getDeleteNextRun() {
		if(deleteTimeInterval.hasValues()) {
			Calendar deleteNextRun = deleteTimeInterval.getNextRun(null); // Parameter=null bedeutet "von jetzt ab"
			if(deleteNextRun != null) {
				return dateFormat.format(deleteNextRun.getTime());
			}
			else {
				return "> 1 Jahr";
			}
		}
		return "---";
	}


	/**
	 * Das Interval als String. Wenn leerer String "", wird der Task nie ausgefhrt.
	 *
	 * @return Das Interval (ausgewertet) als String
	 */
	public String getRequestIntervals() {
		return requestTimeInterval.toString();
	}


	/**
	 * Liefert das Datum, an dem der Task das letzte mal gestartet wurde. Oder einen leeren String, wenn der Task noch nie gestartet wurde.
	 *
	 * @return Datum als String
	 */
	public String getRequestLastRun() {
		if(requestLastRun != null) return dateFormat.format(requestLastRun);
		return "(wurde noch nie gestartet)";
	}


	/**
	 * Liefert das Datum, an dem der Task das nchste mal gestartet wird. Oder einen String, der besagt, dass der Task nie gestartet wird (weil die
	 * konfigurierten Werte ungltig sind). Bercksichtigt werden Daten, die maximal 1 Jahr und 1 Tag in der Zukunft liegen. Tritt kein Fehler auf und wird
	 * trotzdem kein Datum gefunden ist der Rckgabewert "> 1 Jahr".
	 *
	 * @return Datum als String
	 */
	public String getRequestNextRun() {
		if(requestTimeInterval.hasValues()) {
			Calendar requestNextRun = requestTimeInterval.getNextRun(null); // Parameter=null bedeutet "von jetzt ab"
			if(requestNextRun != null) {
				return dateFormat.format(requestNextRun.getTime());
			}
			else {
				return "> 1 Jahr";
			}
		}
		return "---";
	}


	/**
	 * Das Interval (ausgewertet) als String. Wenn leerer String "", wird der Task nie ausgefhrt.
	 *
	 * @return Das Interval (ausgewertet) als String
	 */
	public String getLZVIntervals() {
		if(lzvInterval == 0) {
			return "";
		}
		else if(lzvInterval < (60 * ONE_MINUTE)) {
			return (lzvInterval / ONE_MINUTE) + " Minuten";
		}
		else {
			return (lzvInterval / (60 * ONE_MINUTE)) + " Stunden";
		}
	}


	/**
	 * Liefert das Datum, an dem der Task das letzte mal gestartet wurde. Oder einen leeren String, wenn der Task noch nie gestartet wurde.
	 *
	 * @return Datum als String
	 */
	public String getLZVLastRun() {
		if(lzvLastRun != null) return dateFormat.format(lzvLastRun);
		return "(wurde noch nie gestartet)";
	}


	/** TaskScheduler beenden. Laufende Tasks laufen aber noch zu Ende. Es werden keine neuen Tasks mehr gestartet. */
	public void terminate() {
		terminated = true;
		if(backupCronTask != null) backupCronTask.cancel();
		if(deleteCronTask != null) deleteCronTask.cancel();
		if(requestCronTask != null) requestCronTask.cancel();
		if(lzvTask != null) lzvTask.cancel();
		timer.cancel();
	}


	/**
	 * Private Klasse, die sich um den Aufruf des spontanen Loeschen kmmert. Da dieser Task periodisch gestartet wird, ist er in einer eigenen Klasse
	 * implementiert
	 *
	 * @author Philippe Schrettenbrunner
	 */
	private class IntervalTask extends TimerTask {

		public void run() {
			logger.info("Starte Kapazittscheck");
			try {
				archiveManager.getInQueuesMgr().insertInLZVQueue(LZVTask.getCapacityCheckResultData(archiveManager));
				lzvLastRun = new Date(scheduledExecutionTime());
			}
			catch(RuntimeException e) {
				logger.warning("Fehler beim Start des Kapazittscheck. Der Kapazittscheck (und spontanes Lschen) wird nicht ausgefhrt", e);
			}
		}
	}


	/**
	 * Private Klasse, die sich um den Aufruf von Sichern, Lschen und Nachfordern kmmert. Die Klasse wird mintlich aufgerufen und in der run()-Methode wird
	 * gepfft, ob einer der Tasks gestartet werden muss (auf Grund der Intervallangabe im TimeIntervalCron)
	 *
	 * @author Philippe Schrettenbrunner
	 */
	private class CronTask extends TimerTask {

		private TimeIntervalCron tic;

		public CronTask(TimeIntervalCron tic) {
			this.tic = tic;
		}


		public void run() {
			Date d = new Date(scheduledExecutionTime());
			if(tic.shouldRun(d)) {
				if(tic == backupTimeInterval) {
					logger.info("Starte Sicherungslauf");
					try {
						archiveManager.getInQueuesMgr().insertInBackupQueue(BackupTask.getBackupResultData());
						backupLastRun = d;
					}
					catch(RuntimeException e) {
						logger.warning("Fehler beim Start des Sicherungslaufs", e);
					}
				}
				else if(tic == deleteTimeInterval) {
					logger.info("Starte regulres Loeschen");
					try {
						archiveManager.getInQueuesMgr().insertInDeleteRegularQueue(LZVTask.getDeleteRegularResultData(archiveManager));
						deleteLastRun = d;
					}
					catch(RuntimeException e) {
						logger.warning("Fehler beim Start des des regulren Lschens", e);
					}
				}
				else if(tic == requestTimeInterval) {
					logger.info("Starte Nachforderung fehlender Archivdaten");
					try {
						archiveManager.getInQueuesMgr().insertInRequestQueue(RequestGapTask.getRequestGapResultData());
						requestLastRun = d;
					}
					catch(RuntimeException e) {
						logger.warning("Fehler beim Start der Nachforderung fehlender Archivdaten", e);
					}
				}
			}
		}
	}
}
